# Tests for the datapath module.
#
# Environment variables:
#  DERIVA_PY_TEST_HOSTNAME: hostname of the test server
#  DERIVA_PY_TEST_CREDENTIAL: user credential, if none, it will attempt to get credentail for given hostname
#  DERIVA_PY_TEST_VERBOSE: set for verbose logging output to stdout

from copy import deepcopy
import logging
from operator import itemgetter
import os
import unittest
import sys
from deriva.core import DerivaServer, get_credential, ermrest_model as _em
from deriva.core.datapath import DataPathException, Min, Max, Sum, Avg, Cnt, CntD, Array, ArrayD, Bin

# unittests did not support 'subTests' until 3.4
if sys.version_info[0] < 3 or sys.version_info[1] < 4:
    HAS_SUBTESTS = False
else:
    HAS_SUBTESTS = True

# unittests did not support 'assertWarns' until 3.2
if sys.version_info[0] < 3 or sys.version_info[1] < 2:
    HAS_ASSERTWARNS = False
else:
    HAS_ASSERTWARNS = True

try:
    from pandas import DataFrame
    HAS_PANDAS = True
except ImportError:
    HAS_PANDAS = False

TEST_EXP_MAX = 100
TEST_EXPTYPE_MAX = 10
TEST_EXP_NAME_FORMAT = "experiment-{}"
TEST_PROJ_MAX = 1
TEST_PROJ_INVESTIGATOR = "Smith"
TEST_PROJ_NUM = 1

SPECIAL_CHARACTERS = '`~!@#$%^&*()_+-={}|[]\\;:"\',./<>?'
INVALID_IDENTIFIER, INVALID_IDENTIFIER_FIXED = '9 %$ ', '_9____'
RESERVED_IDENTIFIER = 'column_definitions'
CONFLICTING_IDENTIFIER, CONFLICTING_IDENTIFIER_FIXED = RESERVED_IDENTIFIER + '1', RESERVED_IDENTIFIER + '2'

hostname = os.getenv("DERIVA_PY_TEST_HOSTNAME")
logger = logging.getLogger(__name__)
if os.getenv("DERIVA_PY_TEST_VERBOSE"):
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler())


def define_test_schema(catalog):
    """Defines the test schema.

    A 'vocab' schema with an 'experiment_type' term table.
    An 'isa' schema with an 'experiment' table, with 'type' that references the vocab table.
    """
    model = catalog.getCatalogModel()
    vocab = model.create_schema(_em.Schema.define("Vocab"))
    vocab.create_table(_em.Table.define_vocabulary("Experiment_Type", "TEST:{RID}"))
    isa = model.create_schema(_em.Schema.define("ISA"))

    # create 'Project' table
    table_def = _em.Table.define(
        "Project",
        column_defs=[
            _em.Column.define(cname, ctype) for (cname, ctype) in [
                ('Investigator', _em.builtin_types.text),
                ('Num', _em.builtin_types.int4),
                (INVALID_IDENTIFIER, _em.builtin_types.int4),
                (RESERVED_IDENTIFIER, _em.builtin_types.text),
                (RESERVED_IDENTIFIER + '1', _em.builtin_types.text)
            ]
        ],
        key_defs=[
            _em.Key.define(['Investigator', 'Num'])
        ]
    )
    isa.create_table(table_def)

    # create 'Experiment' table
    table_def = _em.Table.define(
        "Experiment",
        column_defs=[
            _em.Column.define(cname, ctype) for (cname, ctype) in [
                ('Name', _em.builtin_types.text),
                ('Amount', _em.builtin_types.int4),
                ('Time', _em.builtin_types.timestamptz),
                ('Type', _em.builtin_types.text),
                ('Project Investigator', _em.builtin_types.text),
                ('Project_Num', _em.builtin_types.int4),
                ('Empty', _em.builtin_types.int4)
            ]
        ],
        key_defs=[
            _em.Key.define(['Name'])
        ],
        fkey_defs=[
            _em.ForeignKey.define(['Type'], 'Vocab', 'Experiment_Type', ['ID']),
            _em.ForeignKey.define(['Project Investigator', 'Project_Num'], 'ISA', 'Project', ['Investigator', 'Num'])
        ]
    )
    isa.create_table(table_def)

    # create copy of 'Experiment' table
    table_def['table_name'] = 'Experiment_Copy'
    isa.create_table(table_def)


def _generate_experiment_entities(types, count):
    """Generates experiment entities (content only)

    :param types: type entities to be referenced from entities
    :param count: number of entities to return
    :return: a list of dict objects (experiment entities)
    """
    return [
        {
            "Name": TEST_EXP_NAME_FORMAT.format(i),
            "Amount": i,
            "Time": "2018-01-{}T01:00:00.0".format(1 + (i % 31)),
            "Type": types[i % TEST_EXPTYPE_MAX]['ID'],
            "Project Investigator": TEST_PROJ_INVESTIGATOR,
            "Project_Num": TEST_PROJ_NUM,
            "Empty": None
        }
        for i in range(count)
    ]


def populate_test_catalog(catalog):
    """Populate the test catalog."""
    paths = catalog.getPathBuilder()
    logger.debug("Inserting project...")
    logger.debug("Inserting experiment types...")
    proj_table = paths.schemas['ISA'].tables['Project']
    proj_table.insert([
        {"Investigator": TEST_PROJ_INVESTIGATOR, "Num": TEST_PROJ_NUM}
    ])
    type_table = paths.schemas['Vocab'].tables['Experiment_Type']
    types = type_table.insert([
        {"Name": "{}".format(name), "Description": "NA"} for name in range(TEST_EXPTYPE_MAX)
    ], defaults=['ID', 'URI'])
    logger.debug("Inserting experiments...")
    exp = paths.schemas['ISA'].tables['Experiment']
    exp.insert(_generate_experiment_entities(types, TEST_EXP_MAX))


@unittest.skipUnless(hostname, "Test host not specified")
class DatapathTests (unittest.TestCase):
    catalog = None

    @classmethod
    def setUpClass(cls):
        logger.debug("setupUpClass begin")
        credential = os.getenv("DERIVA_PY_TEST_CREDENTIAL") or get_credential(hostname)
        server = DerivaServer('https', hostname, credential)
        cls.catalog = server.create_ermrest_catalog()
        try:
            define_test_schema(cls.catalog)
            populate_test_catalog(cls.catalog)
        except Exception:
            # on failure, delete catalog and re-raise exception
            cls.catalog.delete_ermrest_catalog(really=True)
            raise
        logger.debug("setupUpClass done")

    @classmethod
    def tearDownClass(cls):
        logger.debug("tearDownClass begin")
        cls.catalog.delete_ermrest_catalog(really=True)
        logger.debug("tearDownClass done")

    def setUp(self):
        self.paths = self.catalog.getPathBuilder()
        self.project = self.paths.schemas['ISA'].tables['Project']
        self.experiment = self.paths.schemas['ISA'].tables['Experiment']
        self.experiment_type = self.paths.schemas['Vocab'].tables['Experiment_Type']
        self.experiment_copy = self.paths.schemas['ISA'].tables['Experiment_Copy']
        self.types = list(self.experiment_type.entities())

    def tearDown(self):
        try:
            self.experiment_copy.path.delete()
        except DataPathException:
            # suppresses 404 errors when the table is empty
            pass

    def test_catalog_dir_base(self):
        self.assertIn('schemas', dir(self.paths))

    def test_schema_dir_base(self):
        self.assertLess({'_name', 'tables', 'describe'}, set(dir(self.paths.schemas['ISA'])))

    def test_datapath_dir_base(self):
        self.assertLess({'aggregates', 'groupby', 'attributes', 'context', 'delete', 'entities', 'filter',
                         'link', 'table_instances', 'uri'}, set(dir(self.paths.schemas['ISA'].tables['Experiment'].path)))

    def test_table_dir_base(self):
        self.assertLess({'aggregates', 'alias', 'groupby', 'attributes', 'describe', 'entities', 'filter', 'insert',
                         'link', 'path', 'update', 'uri'}, set(dir(self.paths.schemas['ISA'].tables['Experiment'])))

    def test_catalog_dir_with_schemas(self):
        self.assertLess({'ISA', 'Vocab'}, set(dir(self.paths)))

    def test_schema_dir_with_tables(self):
        self.assertIn('Experiment', dir(self.paths.ISA))

    def test_table_dir_with_columns(self):
        self.assertLess({'Name', 'Amount', 'Time', 'Type'}, set(dir(self.paths.ISA.Experiment)))

    def test_dir_path(self):
        self.assertIn('Experiment', dir(self.paths.ISA.Experiment.path))

    def test_dir_invalid_identifier(self):
        self.assertIn(INVALID_IDENTIFIER_FIXED, dir(self.project))
        self.assertIsNotNone(getattr(self.project, INVALID_IDENTIFIER_FIXED))

    def test_dir_conflicting_identifier(self):
        self.assertIn(CONFLICTING_IDENTIFIER_FIXED, dir(self.project))
        self.assertIsNotNone(getattr(self.project, CONFLICTING_IDENTIFIER))
        self.assertIsNotNone(getattr(self.project, CONFLICTING_IDENTIFIER_FIXED))

    def test_describe_schema(self):
        with self.assertWarns(DeprecationWarning):
            self.paths.schemas['ISA'].describe()

    def test_describe_table(self):
        with self.assertWarns(DeprecationWarning):
            self.paths.schemas['ISA'].tables['Experiment'].describe()

    def test_describe_column(self):
        with self.assertWarns(DeprecationWarning):
            self.paths.schemas['ISA'].tables['Experiment'].column_definitions['Name'].describe()

    def test_unfiltered_fetch(self):
        results = self.experiment.entities()
        self.assertEqual(len(results), TEST_EXP_MAX)

    def test_fetch_with_limit(self):
        results = self.experiment.entities()
        limit = TEST_EXP_MAX / 5
        results.fetch(limit=limit)
        self.assertEqual(len(results), limit)

    def test_fetch_with_sort(self):
        results = self.experiment.entities()
        results.sort(self.experiment.column_definitions['Amount'])
        self.assertEqual(results[0]['Amount'], 0)

    def test_fetch_attributes_with_sort(self):
        results = self.experiment.attributes(self.experiment.RID, self.experiment.Amount)
        results.sort(self.experiment.Amount)
        self.assertEqual(results[0]['Amount'], 0)

    def test_fetch_all_attributes_with_sort(self):
        results = self.experiment.attributes(self.experiment)
        results.sort(self.experiment.Amount)
        self.assertEqual(results[0]['Amount'], 0)

    def test_fetch_all_attributes_with_sort_desc(self):
        results = self.experiment.attributes(self.experiment)
        results.sort(self.experiment.Amount.desc)
        self.assertEqual(results[0]['Amount'], TEST_EXP_MAX-1)

    def test_fetch_from_path_attributes_with_sort_on_talias(self):
        path = self.experiment.path
        results = path.Experiment.attributes(path.Experiment.RID, path.Experiment.Amount)
        results.sort(path.Experiment.Amount)
        self.assertEqual(results[0]['Amount'], 0)

    def test_fetch_from_path_attributes_with_sort_on_talias_desc(self):
        path = self.experiment.path
        results = path.Experiment.attributes(path.Experiment.RID, path.Experiment.Amount)
        results.sort(path.Experiment.Amount.desc)
        self.assertEqual(results[0]['Amount'], TEST_EXP_MAX-1)

    def test_fetch_from_path_all_attributes_with_sort_on_talias(self):
        path = self.experiment.path
        results = path.Experiment.attributes(*path.Experiment.column_definitions.values())
        results.sort(path.Experiment.Amount)
        self.assertEqual(results[0]['Amount'], 0)

    def test_fetch_from_path_all_attributes_with_sort_on_alias_desc(self):
        path = self.experiment.path
        results = path.Experiment.attributes(*path.Experiment.column_definitions.values())
        results.sort(path.Experiment.Amount.desc)
        self.assertEqual(results[0]['Amount'], TEST_EXP_MAX-1)

    def test_attribute_projection(self):
        results = self.experiment.attributes(
            self.experiment.column_definitions['Name'],
            self.experiment.column_definitions['Amount']
        )
        result = results.fetch(limit=1)[0]
        self.assertIn('Name', result)
        self.assertIn('Amount', result)

    def test_attribute_err_table_attr(self):
        table_attr = ['_name', '_schema']
        for attr in table_attr:
            with self.assertRaises(TypeError):
                self.experiment.attributes(getattr(self.experiment, attr))

    def test_update_err_no_targets(self):
        entities = [{'RID': 1234}]
        with self.assertRaises(ValueError):
            self.experiment.update(entities)

    def test_aggregate_w_invalid_attributes(self):
        with self.assertRaises(TypeError):
            self.experiment.aggregates(Min(self.experiment.column_definitions['Amount']))

    def test_aggregate_w_invalid_renames(self):
        with self.assertRaises(TypeError):
            self.experiment.aggregates(
                self.experiment.column_definitions['Name'],
                Min(self.experiment.column_definitions['Amount'])
            )

    @unittest.skipUnless(HAS_SUBTESTS, "This tests is not available unless running python 3.4+")
    def test_aggregate_fns(self):
        tests = [
            ('min_amount',      Min,    0),
            ('max_amount',      Max,    TEST_EXP_MAX-1),
            ('sum_amount',      Sum,    sum(range(TEST_EXP_MAX))),
            ('avg_amount',      Avg,    sum(range(TEST_EXP_MAX))/TEST_EXP_MAX),
            ('cnt_amount',      Cnt,    TEST_EXP_MAX),
            ('cnt_d_amount',    CntD,   TEST_EXP_MAX),
            ('array_amount',    Array,  list(range(TEST_EXP_MAX))),
            ('array_d_amount',  ArrayD, list(range(TEST_EXP_MAX)))
        ]
        for name, Fn, value in tests:
            with self.subTest(name=name):
                # results = self.experiment.aggregates(**{name: Fn(self.experiment.column_definitions['Amount'])})
                results = self.experiment.aggregates(Fn(self.experiment.column_definitions['Amount']).alias(name))
                result = results.fetch()[0]
                self.assertIn(name, result)
                self.assertEqual(result[name], value)

    def test_aggregate_w_2_fns(self):
        results = self.experiment.aggregates(
            Min(self.experiment.column_definitions['Amount']).alias('min_amount'),
            Max(self.experiment.column_definitions['Amount']).alias('max_amount')
        )
        result = results.fetch()[0]
        self.assertIn('min_amount', result)
        self.assertEqual(result['min_amount'], 0)
        self.assertIn('max_amount', result)
        self.assertEqual(result['max_amount'], TEST_EXP_MAX-1)

    @unittest.skipUnless(HAS_SUBTESTS, "This tests is not available unless running python 3.4+")
    def test_aggregate_fns_array_star(self):
        path = self.experiment.path
        tests = [
            ('array_table_star',  Array,  self.experiment, self.experiment),
            ('array_alias_star',  Array,  path,            path.Experiment),
            ('arrayd_table_star', ArrayD, self.experiment, self.experiment),
            ('arrayd_alias_star', ArrayD, path,            path.Experiment)
        ]
        for name, Fn, path, instance in tests:
            results = path.aggregates(Fn(instance).alias('arr'))
            with self.subTest(name=name):
                result = results.fetch()[0]
                self.assertIn('arr', result)
                self.assertEqual(len(result['arr']), TEST_EXP_MAX)
                self.assertIn('Time', result['arr'][0])

    @unittest.skipUnless(HAS_SUBTESTS, "This tests is not available unless running python 3.4+")
    def test_aggregate_fns_cnt_star(self):
        path = self.experiment.path
        tests = [
            ('cnt_table_star', Cnt, self.experiment, self.experiment),
            ('cnt_alias_star', Cnt, path,            path.Experiment)
        ]
        for name, Fn, path, instance in tests:
            results = path.aggregates(Fn(instance).alias('cnt'))
            with self.subTest(name=name):
                result = results.fetch()[0]
                self.assertIn('cnt', result)
                self.assertEqual(result['cnt'], TEST_EXP_MAX)

    @unittest.skipUnless(HAS_SUBTESTS, "This tests is not available unless running python 3.4+")
    def test_attributegroup_fns(self):
        tests = [
            ('one group key',     [self.experiment.column_definitions['Type']]),
            ('two group keys',    [self.experiment.column_definitions['Project_Num'], self.experiment.column_definitions['Type']]),
            ('aliased group key', [self.experiment.column_definitions['Type'].alias('The Type')])
        ]
        for test_name, group_key in tests:
            with self.subTest(name=test_name):
                self._do_attributegroup_fn_subtests(group_key)

    @unittest.skipUnless(HAS_SUBTESTS, "This tests is not available unless running python 3.4+")
    def _do_attributegroup_fn_subtests(self, group_key):
        """Helper method for running common attributegroup subtests for different group keys."""
        tests = [
            ('min_amount',      Min,    0),
            ('max_amount',      Max,    TEST_EXP_MAX-TEST_EXPTYPE_MAX),
            ('sum_amount',      Sum,    sum(range(0, TEST_EXP_MAX, TEST_EXPTYPE_MAX))),
            ('avg_amount',      Avg,    sum(range(0, TEST_EXP_MAX, TEST_EXPTYPE_MAX))/TEST_EXPTYPE_MAX),
            ('cnt_amount',      Cnt,    TEST_EXPTYPE_MAX),
            ('cnt_d_amount',    CntD,   TEST_EXPTYPE_MAX),
            ('array_amount',    Array,  list(range(0, TEST_EXP_MAX, TEST_EXPTYPE_MAX))),
            ('array_d_amount',  ArrayD, list(range(0, TEST_EXP_MAX, TEST_EXPTYPE_MAX)))
        ]
        for name, Fn, value in tests:
            with self.subTest(name=name):
                results = self.experiment.groupby(*group_key).attributes(
                    Fn(self.experiment.column_definitions['Amount']).alias(name)).sort(*group_key)

                result = results[0]
                self.assertEqual(len(results), TEST_EXPTYPE_MAX)
                self.assertTrue(all(key._name in result for key in group_key))
                self.assertIn(name, result)
                self.assertEqual(result[name], value)

    @unittest.skipUnless(HAS_SUBTESTS, "This tests is not available unless running python 3.4+")
    def test_attributegroup_w_bin(self):
        tests = [
            ('min/max given',     0,    TEST_EXP_MAX),
            ('min/max not given', None, None),
            ('min only given',    0,    None),
            ('max only given',    None, TEST_EXP_MAX)
        ]
        for testname, minval, maxval in tests:
            with self.subTest(name=testname):
                self._do_bin_subtests(minval, maxval)

    def _do_bin_subtests(self, minval, maxval):
        """Helper method for running common binning tests with & without min/max values."""
        new_name, bin_name = 'TheProj', 'ABin'
        nbins = int(TEST_EXP_MAX/20)
        group_key = [
            self.experiment.column_definitions['Project_Num'].alias(new_name),
            Bin(self.experiment.column_definitions['Amount'], nbins, minval=minval, maxval=maxval).alias(bin_name)
        ]
        tests = [
            ('min_amount',      Min,    lambda a, b: a >= b[1]),
            ('max_amount',      Max,    lambda a, b: a <= b[2]),
            ('sum_amount',      Sum,    lambda a, b: a >= b[1] + b[2]),
            ('avg_amount',      Avg,    lambda a, b: b[1] <= a <= b[2]),
            ('cnt_amount',      Cnt,    lambda a, b: a == TEST_EXP_MAX/nbins),
            ('cnt_d_amount',    CntD,   lambda a, b: a == TEST_EXP_MAX/nbins),
            ('array_amount',    Array,  lambda a, b: all(b[1] <= a_i <= b[2] for a_i in a)),
            ('array_d_amount',  ArrayD, lambda a, b: all(b[1] <= a_i <= b[2] for a_i in a))
        ]
        for name, Fn, compare in tests:
            with self.subTest(name=name):
                results = self.experiment.groupby(*group_key).attributes(
                    Fn(self.experiment.column_definitions['Amount']).alias(name)).fetch()

                self.assertTrue(all(key._name in results[0] for key in group_key))
                self.assertIn(name, results[0])
                for result in results:
                    bin = result[bin_name]
                    if not maxval and (bin[0] >= nbins):
                        # skip the last 2 bins when maxval was resolved; those bins are not aligned like the others
                        continue
                    self.assertTrue(compare(result[name], bin))

    @unittest.skipUnless(HAS_SUBTESTS, "This tests is not available unless running python 3.4+")
    def test_attributegroup_w_bin_sort(self):
        bin_name = 'bin'
        nbins = int(TEST_EXP_MAX/20)
        bin = Bin(self.experiment.column_definitions['Amount'], nbins, 0, TEST_EXP_MAX).alias(bin_name)
        bin_desc = bin.desc
        asc_fn = lambda n, a, b: a[n] <= b[n]
        desc_fn = lambda n, a, b: a[n] >= b[n]
        tests = [
            ('min_amount', Min, bin,      asc_fn),
            ('max_amount', Max, bin,      asc_fn),
            ('sum_amount', Sum, bin,      asc_fn),
            ('avg_amount', Avg, bin,      asc_fn),
            ('min_amount', Min, bin_desc, desc_fn),
            ('max_amount', Max, bin_desc, desc_fn),
            ('sum_amount', Sum, bin_desc, desc_fn),
            ('avg_amount', Avg, bin_desc, desc_fn)
        ]
        for name, Fn, sort_key, compfn in tests:
            with self.subTest(name=name):
                results = self.experiment.groupby(bin).attributes(
                    Fn(self.experiment.column_definitions['Amount']).alias(name)).sort(sort_key).fetch()

                self.assertIn(bin._name, results[0])
                self.assertIn(name, results[0])
                self.assertTrue(compfn(name, results[0], results[1]))

    @unittest.skipUnless(HAS_SUBTESTS, "This test is not available unless running python 3.4+")
    def test_attributegroup_w_bin_resolution(self):
        binkey = self.experiment.column_definitions['Empty']
        binname = 'bin'
        tests = [
            ('min_max_valid', 0,    0,      True),
            ('max_invalid',   0,    None,   False),
            ('min_invalid',   None, 0,      False),
            ('both_invalid',  None, None,   False)
        ]
        for name, minval, maxval, valid in tests:
            def _do_query():
                bin = Bin(binkey, 10, minval, maxval).alias(binname)
                return self.experiment.groupby(bin).attributes(Avg(binkey).alias(name)).fetch()

            with self.subTest(name=name):
                if valid:
                    results = _do_query()
                    self.assertIn(binname, results[0])
                    self.assertIn(name, results[0])
                else:
                    with self.assertRaises(ValueError):
                        _do_query()

    def test_link_implicit(self):
        results = self.experiment.link(self.experiment_type).entities()
        self.assertEqual(TEST_EXPTYPE_MAX, len(results))

    def test_link_explicit_simple_key(self):
        results = self.experiment.link(
            self.experiment_type,
            on=(self.experiment.Type == self.experiment_type.ID)
        ).entities()
        self.assertEqual(TEST_EXPTYPE_MAX, len(results))

    def test_link_explicit_composite_key(self):
        path = self.experiment.link(
            self.project,
            on=(
                    (self.experiment.Project_Investigator == self.project.Investigator) &
                    (self.experiment.Project_Num == self.project.Num)
            )
        )
        results = path.entities()
        self.assertEqual(TEST_PROJ_MAX, len(results))

    def test_filter_equality(self):
        results = self.experiment.filter(
            self.experiment.column_definitions['Name'] == TEST_EXP_NAME_FORMAT.format(1)
        ).entities()
        self.assertEqual(len(results), 1)

    def test_filter_inequality(self):
        results = self.experiment.filter(
            self.experiment.column_definitions['Amount'] < 10
        ).entities()
        self.assertEqual(len(results), 10)

    def test_filter_ciregexp(self):
        results = self.experiment.filter(
            self.experiment.column_definitions['Name'].ciregexp(TEST_EXP_NAME_FORMAT.format(0)[10:])
        ).entities()
        self.assertEqual(len(results), 1)

    def test_filter_negation(self):
        results = self.experiment.filter(
            ~ (self.experiment.column_definitions['Name'].ciregexp(TEST_EXP_NAME_FORMAT.format(0)[10:]))
        ).entities()
        self.assertEqual(len(results), TEST_EXP_MAX - 1)

    def test_filter_conjunction(self):
        results = self.experiment.filter(
            self.experiment.column_definitions['Name'].ciregexp(TEST_EXP_NAME_FORMAT.format(0)[10:])
            & (self.experiment.column_definitions['Amount'] == 0)
        ).entities()
        self.assertEqual(len(results), 1)

    def test_attribute_deprecated_rename(self):
        with self.assertRaises(TypeError):
            self.experiment.attributes(
                self.experiment.column_definitions['Name'],
                howmuch=self.experiment.column_definitions['Amount']
            )

    def test_attribute_rename(self):
        results = self.experiment.attributes(
            self.experiment.column_definitions['Name'],
            self.experiment.column_definitions['Amount'].alias('How many of them ?'),
            self.experiment.column_definitions['Project_Num'].alias('Project #')
        )
        result = results.fetch(limit=1)[0]
        self.assertIn('Name', result)
        self.assertIn('How many of them ?', result)
        self.assertIn('Project #', result)

    def test_attribute_rename_special_chars(self):
        # first test with only the `:` character present which would trigger a lexical error from ermrest
        special_character_out_alias = self.experiment._name + ':' + self.experiment.column_definitions['Name']._name
        results = self.experiment.attributes(self.experiment.column_definitions['Name'].alias(special_character_out_alias))
        result = results.fetch(limit=1)[0]
        self.assertIn(special_character_out_alias, result)

        # second test with url unsafe characters present which would trigger a bad request from the web server
        special_character_out_alias = SPECIAL_CHARACTERS
        results = self.experiment.attributes(self.experiment.column_definitions['Name'].alias(special_character_out_alias))
        result = results.fetch(limit=1)[0]
        self.assertIn(special_character_out_alias, result)

    def test_context(self):
        path = self.experiment.link(self.experiment_type)
        results = path.Experiment.entities()
        self.assertEqual(len(results), TEST_EXP_MAX)

    def test_path_table_instances(self):
        path = self.experiment.link(self.experiment_type)
        results = path.table_instances['Experiment'].entities()
        self.assertEqual(len(results), TEST_EXP_MAX)

    def test_path_project(self):
        path = self.experiment.link(self.experiment_type)
        results = path.Experiment.attributes(
            path.Experiment,
            path.Experiment_Type.column_definitions['URI'],
            path.Experiment_Type.column_definitions['Name'].alias('exptype')
        )
        result = results.fetch(limit=1)[0]
        self.assertIn('Experiment:Name', result)
        self.assertIn('Experiment:Time', result)
        self.assertIn('URI', result)
        self.assertIn('exptype', result)

    @unittest.skipUnless(HAS_PANDAS, "pandas library not available")
    def test_dataframe(self):
        results = self.experiment.entities()
        df = DataFrame(results)
        self.assertEqual(len(df), TEST_EXP_MAX)

    def test_insert_empty_entities(self):
        results = self.experiment_copy.insert(None)
        self.assertEqual(len(results), 0)
        results = self.experiment_copy.insert([])
        self.assertEqual(len(results), 0)

    def test_insert_entities_not_iterable(self):
        with self.assertRaises(TypeError):
            self.experiment_type.insert(1)

    def test_insert_entities0_not_dict(self):
        with self.assertRaises(TypeError):
            self.experiment_type.insert([1])
        with self.assertRaises(TypeError):
            self.experiment_type.insert('this is not a dict')

    def test_insert(self):
        results = self.experiment_copy.insert(_generate_experiment_entities(self.types, 10))
        self.assertEqual(len(results), 10)

    def test_update(self):
        inserted = self.experiment_copy.insert(_generate_experiment_entities(self.types, 10))
        self.assertEqual(len(inserted), 10)
        # now change something in the first result
        updates = [dict(**inserted[0])]
        updates[0]['Name'] = '**CHANGED**'
        updated = self.experiment_copy.update(updates)
        self.assertEqual(len(updated), 1)
        self.assertEqual(inserted[0]['RID'], updated[0]['RID'])
        self.assertNotEqual(inserted[0]['Name'], updated[0]['Name'])

    def test_update_empty_entities(self):
        results = self.experiment_copy.update(None)
        self.assertEqual(len(results), 0)
        results = self.experiment_copy.update([])
        self.assertEqual(len(results), 0)

    def test_update_entities_not_iterable(self):
        with self.assertRaises(TypeError):
            self.experiment_type.update(1)

    def test_update_entities0_not_dict(self):
        with self.assertRaises(TypeError):
            self.experiment_type.update([1])
        with self.assertRaises(TypeError):
            self.experiment_type.update('this is not a dict')

    def test_nondefaults(self):
        nondefaults = {'RID', 'RCB', 'RCT'}
        results = self.experiment.entities()
        self.assertEqual(len(results), TEST_EXP_MAX)
        entities_copy = self.experiment_copy.insert(results, nondefaults=nondefaults, add_system_defaults=False)
        self.assertEqual(len(results), len(entities_copy), 'entities not copied completely')
        ig = itemgetter(*nondefaults)
        for i in range(TEST_EXP_MAX):
            self.assertEqual(ig(results[i]), ig(entities_copy[i]), 'copied values do not match')

    def test_nondefaults_w_add_sys_defaults(self):
        nondefaults = {'RID', 'RCB', 'RCT'}
        results = self.experiment.entities()
        self.assertEqual(len(results), TEST_EXP_MAX)
        entities_copy = self.experiment_copy.insert(results, nondefaults=nondefaults)
        self.assertEqual(len(results), len(entities_copy), 'entities not copied completely')
        ig = itemgetter(*nondefaults)
        for i in range(TEST_EXP_MAX):
            self.assertEqual(ig(results[i]), ig(entities_copy[i]), 'copied values do not match')

    @unittest.skipUnless(HAS_SUBTESTS, "This test is not available unless running python 3.4+")
    def test_deepcopy_of_paths(self):
        paths = [
            self.experiment.path,
            self.experiment.link(self.experiment_type),
            self.experiment.link(self.experiment_type, on=(self.experiment.Type == self.experiment_type.ID)),
            self.experiment.link(
                self.project,
                on=(
                        (self.experiment.Project_Investigator == self.project.Investigator) &
                        (self.experiment.Project_Num == self.project.Num)
                )
            ),
            self.project.filter(self.project.Num < 1000).link(self.experiment).link(self.experiment_type),
            self.experiment.alias('Exp').link(self.experiment_type.alias('ExpType')),
            self.experiment.filter(self.experiment.column_definitions['Name'] == TEST_EXP_NAME_FORMAT.format(1)),
            self.experiment.filter(self.experiment.column_definitions['Amount'] < 10),
            self.experiment.filter(
                self.experiment.column_definitions['Name'].ciregexp(TEST_EXP_NAME_FORMAT.format(0)[10:])
            ),
            self.experiment.filter(
                ~ (self.experiment.column_definitions['Name'].ciregexp(TEST_EXP_NAME_FORMAT.format(0)[10:]))
            ),
            self.experiment.filter(
                self.experiment.column_definitions['Name'].ciregexp(TEST_EXP_NAME_FORMAT.format(0)[10:])
                & (self.experiment.column_definitions['Amount'] == 0)
            )
        ]
        for path in paths:
            with self.subTest(name=path.uri):
                cp = deepcopy(path)
                self.assertNotEqual(path, cp)
                self.assertEqual(path.uri, cp.uri)

    def test_merge_paths(self):
        path1 = self.experiment.filter(self.experiment.Amount >= 0)
        path2 = self.experiment.link(self.experiment_type).filter(self.experiment_type.ID >= '0')
        path3 = self.experiment.link(self.project).filter(self.project.Num >= 0)
        original_uri = path1.uri

        # merge paths 1..3
        path1.merge(path2).merge(path3)
        self.assertNotEqual(path1.uri, original_uri, "Merged path's URI should have changed from its original URI")
        self.assertEqual(path1.context._name, path3.context._name, "Context of merged paths should equal far right-hand path's context")
        self.assertGreater(len(path1.Experiment.entities()), 0, "Should have returned results")

    def test_compose_paths(self):
        path1 = self.experiment.filter(self.experiment.Amount >= 0)
        path2 = self.experiment.link(self.experiment_type).filter(self.experiment_type.ID >= '0')
        path3 = self.experiment.link(self.project).filter(self.project.Num >= 0)
        original_uri = path1.uri

        # compose paths 1..3
        path = self.paths.compose(path1, path2, path3)
        self.assertNotEqual(path, path1, "Compose should have copied the first path rather than mutate it")
        self.assertNotEqual(path.uri, path1.uri, "Composed path URI should not match the first path URI")
        self.assertEqual(path1.uri, original_uri, "First path was changed")
        self.assertNotEqual(path.uri, original_uri, "Merged path's URI should have changed from its original URI")
        self.assertEqual(path.context._name, path3.context._name, "Context of composed paths should equal far right-hand path's context")
        self.assertGreater(len(path.Experiment.entities()), 0, "Should have returned results")


if __name__ == '__main__':
    sys.exit(unittest.main())



import os
import json
import importlib
import logging
import re
from copy import copy
from enum import Enum
from deriva.transfer.download import DerivaDownloadError, DerivaDownloadConfigurationError
from deriva.transfer.download.processors.transform.base_transform_processor import BaseTransformProcessor, \
    LOCAL_PATH_KEY, SOURCE_URL_KEY


# this processor transforms sequence data in RBK to excel file of GEO format
# https://www.ncbi.nlm.nih.gov/geo/info/seq.html
class GeoExportTransformProcessor(BaseTransformProcessor):
    def __init__(self, envars=None, **kwargs):
        super(GeoExportTransformProcessor, self).__init__(envars, **kwargs)
        self._create_input_output_paths()

    def process(self):
        logging.info("Transforming input file [%s] into output file [%s]" %
                     (self.input_abspath, self.output_abspath))
        # input_abspath: the json file containing raw data retrieved from DB
        # output_abspath: the excel file of GEO format
        geo = Export2GEO(self.input_abspath, self.output_abspath, self.parameters)
        geo.export_all()

        super(GeoExportTransformProcessor, self).process()
        return self.outputs


# used for setting background color
class Style(Enum):
    DEFAULT = 0
    INSTRUCTION = 1
    SECTION = 2
    FIELD = 3
    HEADER = 4
    PREFACE = 5


# handling excel operations
class Export2Excel(object):
    OPENPYXL = None

    def __init__(self, output_path):
        self.__name__ = 'export_excel'
        self.wb = None
        self.ws = None
        self._output_path_ = output_path
        self.sheet_name = 'METADATA'
        self.initialize_wb()

    def import_openpyxl(self):
        # locate library
        if self.OPENPYXL is None:
            try:
                self.OPENPYXL = importlib.import_module("openpyxl")
            except ImportError as e:
                raise DerivaDownloadConfigurationError("Unable to find required module. "
                                                       "Ensure that the Python package \"openpyxl\" is installed.", e)

    def initialize_wb(self):
        self.import_openpyxl()
        # Open an xlsx
        self.wb = self.OPENPYXL.Workbook()
        self.ws = self.wb.create_sheet(index=0, title=self.sheet_name)

        # remove the default sheet
        self.wb.remove(self.wb['Sheet'])

        fgColor0 = self.OPENPYXL.styles.colors.Color(indexed=10, type='indexed')
        fgColor1 = self.OPENPYXL.styles.colors.Color(indexed=42, type='indexed')
        bgColor1 = self.OPENPYXL.styles.colors.Color(indexed=27, type='indexed')
        fgColor2 = self.OPENPYXL.styles.colors.Color(indexed=12, type='indexed')
        fgColor3 = self.OPENPYXL.styles.colors.Color(indexed=13, type='indexed')
        bgColor3 = self.OPENPYXL.styles.colors.Color(indexed=34, type='indexed')

        self.font_red = self.OPENPYXL.styles.Font(name='Arial', b=True, color=fgColor0, sz=10.0)
        self.font_blue = self.OPENPYXL.styles.Font(name='Arial', b=True, color=fgColor2, sz=10.0)
        self.font_bond = self.OPENPYXL.styles.Font(name='Arial', b=True, sz=10.0)

        self.fill_green = self.OPENPYXL.styles.PatternFill('solid', fgColor=fgColor1, bgColor=bgColor1)
        self.fill_yellow = self.OPENPYXL.styles.PatternFill('solid', fgColor=fgColor3, bgColor=bgColor3)

        self.ws.column_dimensions['A'].width = 42.0
        self.ws.column_dimensions['B'].width = 15.0
        self.ws.column_dimensions['C'].width = 24.0
        self.ws.column_dimensions['D'].width = 15.0
        self.ws.column_dimensions['E'].width = 24.0
        self.ws.column_dimensions['F'].width = 24.0
        self.ws.column_dimensions['G'].width = 24.0
        self.ws.column_dimensions['H'].width = 24.0
        self.ws.column_dimensions['I'].width = 24.0
        self.ws.column_dimensions['J'].width = 24.0
        self.ws.column_dimensions['K'].width = 24.0
        self.ws.column_dimensions['L'].width = 24.0
        self.ws.column_dimensions['M'].width = 24.0
        self.ws.column_dimensions['N'].width = 24.0

    @staticmethod
    # row id and column id to excel cell location
    def cell_name(row_num, col_num):
        string = ""
        while col_num > 0:
            col_num, remainder = divmod(col_num - 1, 26)
            string = chr(65 + remainder) + string
        return string + str(row_num)

    def insert_row(self, row_idx, row_amt):
        self.ws.insert_rows(idx=row_idx, amount=row_amt)

    def delete_row(self, idx, row_amt):
        self.ws.delete_rows(idx=idx, amount=row_amt)

    # writing value into cell with some style
    def write_cell(self, row_idx, col_idx, content, export_style=Style.DEFAULT):
        cell_loc = self.cell_name(row_idx, col_idx)
        cell = self.ws[cell_loc]
        cell.value = content
        if (export_style == Style.INSTRUCTION):
            cell.fill = self.fill_green
        elif (export_style == Style.HEADER):
            cell.fill = self.fill_yellow
            cell.font = self.font_blue
        elif (export_style == Style.FIELD):
            cell.font = self.font_blue
        elif (export_style == Style.SECTION):
            cell.font = self.font_red
        elif (export_style == Style.PREFACE):
            cell.fill = self.fill_green
            cell.font = self.font_bond
        else:
            pass

    def save_xlsx(self):
        self.wb.save(self._output_path_)

    def copy_cell(self, old_row_idx, old_col_idx, new_row_idx, new_col_idx, copy_style=True):
        cell = self.ws[self.cell_name(old_row_idx, old_col_idx)]
        new_cell = self.ws[self.cell_name(new_row_idx, new_col_idx)]
        new_cell.value = cell.value
        if cell.has_style and copy_style:
            new_cell.font = copy(cell.font)
            new_cell.border = copy(cell.border)
            new_cell.fill = copy(cell.fill)
            new_cell.number_format = copy(cell.number_format)
            new_cell.protection = copy(cell.protection)
            new_cell.alignment = copy(cell.alignment)
            new_cell.comment = copy(cell.comment)


class Export2GEO(object):

    def __init__(self, input_file, output_excel_path, parameters):
        self.__name__ = 'export_geo_template'
        self.input_file = input_file
        self.output_excel_path = output_excel_path
        self.parameters = parameters
        self.current_row_idx = 0
        self.header_row_idx = 0
        # all types of protocols
        self.protocol_type = ['Isolation_Protocol', 'Growth_Protocol', 'Treatment_Protocol', 'Extract_Protocol',
                              'Construction_Protocol', 'Label_Protocol', 'Hybridization_Protocol', 'Scan_Protocol']
        self.protocol_unique = {}
        self.protocol_list = {}
        self.other_item = ['Data_Processing', 'Reference_Genome']
        self.other_item_unique = {}
        self.other_item_list = {}
        self.initialize_data()

    def initialize_data(self):
        with open(self.input_file) as input_file:
            data = json.load(input_file)
            self.study = None
            if data is None or len(data) == 0:
                logging.debug("No Data")
            elif len(data) > 1:
                logging.debug("Duplicate Data")
            elif len(data) == 1:
                # get parameters from json file
                self.study = data[0].get('Study')[0]
                self.pi = data[0].get('Principal_Investigator')[0]
                self.experiments = data[0].get('Experiment')
                self.experiment_settings = data[0].get('Experiment_Setting')
                self.replicates = data[0].get('Replicate')
                self.specimen = data[0].get('Specimen')
                self.files = data[0].get('File')
                self.study_files = data[0].get('Study_File')
                self.stage = data[0].get('Stage')
                self.tissue = data[0].get('Tissue')
                self.source_type = data[0].get('Source')
                self.specimen_cell_type = data[0].get('Specimen_Cell_Type')
                self.cell_type = data[0].get('Cell_Type')
                self.specimen_allele = data[0].get('Specimen_Allele')
                self.allele = data[0].get('Allele')


                # Adding sorting to files based on key ( Replicate-RID, FileName )
                # https://github.com/informatics-isi-edu/rbk-project/issues/615
                if self.replicates and len(self.replicates) > 1 and self.replicates[0] is not None:
                    self.replicates.sort( key = lambda x: ( x.get('RID') ) )
                
                if self.files and len(self.files) > 1 and self.files[0] is not None:
                    self.files.sort( key = lambda x: ( x.get('Replicate_RID'), x.get('File_Name') ) )

                if self.study_files and len(self.study_files) > 1 and self.study_files[0] is not None:
                    self.study_files.sort( key = lambda x: ( x.get('File_Name') ) )



    def export_start(self):
        self.excel = Export2Excel(self.output_excel_path)

    def export_all(self):
        if self.study is not None:
            # export data for different sections such as series,samples etc.
            self.export_start()
            self.export_preface()
            self.export_series()
            self.export_sample_prepare()
            self.export_sample()
            self.export_protocol()
            self.export_data_processing()
            self.export_processed_datafiles()
            self.export_raw_datafiles()
            self.export_paired_end()
            self.export_finish()

    # export preface sections in GEO template excel
    # words may need update if GEO template update
    # https://www.ncbi.nlm.nih.gov/geo/info/seq.html#metadata
    def export_preface(self):
        self.current_row_idx = 1
        self.excel.write_cell(self.current_row_idx, 1, '# High-throughput sequencing metadata template (version 2.1).',
                              Style.PREFACE)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, '# All fields in this template must be completed.',
                              Style.PREFACE)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1,
                              '# Templates containing example data are found in the METADATA EXAMPLES spreadsheet tabs at the foot of this page.',
                              Style.PREFACE)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1,
                              '# Field names (in blue on this page) should not be edited. Hover over cells containing field names to view field content guidelines.',
                              Style.PREFACE)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, (
            '# Human data. If there are patient privacy concerns regarding making data fully public through GEO, '
            'please submit to NCBI''s dbGaP (http://www.ncbi.nlm.nih.gov/gap/) database. dbGaP has controlled access'
            'mechanisms and is an appropriate resource for hosting sensitive patient data.'), Style.PREFACE)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, '', Style.PREFACE)
        self.current_row_idx += 1

        for col in range(2, 13):
            for row in range(1, 7):
                self.excel.write_cell(row, col, '', Style.PREFACE)

    # export SERIES(study)
    def export_series(self):
        self.excel.write_cell(self.current_row_idx, 1, 'SERIES', Style.SECTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, '# This section describes the overall experiment.',
                              Style.INSTRUCTION)
        for col in range(2, 13):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'title', Style.FIELD)
        self.excel.write_cell(self.current_row_idx, 2, self.study.get('Title', ''))
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'summary', Style.FIELD)
        self.excel.write_cell(self.current_row_idx, 2, self.study.get('Summary', ''))
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'overall design', Style.FIELD)
        self.excel.write_cell(self.current_row_idx, 2, self.study.get('Overall_Design', ''))
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'contributor', Style.FIELD)
        self.excel.write_cell(self.current_row_idx, 2, self.pi.get('Full_Name', ''))
        self.current_row_idx += 1

        # one line for each study file
        if len(self.study_files) >= 1 and self.study_files[0] is not None:
            file_path = self.parameters.get("study_files_path_template", self.study_files[0].get("File_Name", ""))
            self.excel.write_cell(self.current_row_idx, 1, 'supplementary file', Style.FIELD)
            self.excel.write_cell(self.current_row_idx, 2, file_path.format(**self.study_files[0]))
            self.current_row_idx += 1
            for i in range(len(self.study_files) - 1):
                file_path = self.parameters.get("study_files_path_template", self.study_files[i + 1].get("File_Name", ""))
                self.excel.write_cell(self.current_row_idx, 1, 'supplementary file', Style.FIELD)
                self.excel.write_cell(self.current_row_idx, 2,  file_path.format(**self.study_files[i + 1]))
                self.current_row_idx += 1
        else:
            self.excel.write_cell(self.current_row_idx, 2, '')
            logging.debug('No study file')

    # prepare the data for sample export
    def export_sample_prepare(self):
        # check self.protocol unique
        for p in self.protocol_type:
            # protocols which are applicable to only a subset of Samples should be included
            # as additional columns of the SAMPLES section instead.
            # this dictionary stores if containing unique content for each protocol type or not
            # used to handle dynamic headers related to the protocols
            self.protocol_unique[p] = True
            # initialize list of items for each protocol type
            self.protocol_list[p] = []

        for i in self.other_item:
            self.other_item_unique[i] = True
            self.other_item_list[i] = []

        for e in self.experiments:
            if e is None or 'RID' not in e.keys():
                continue
            for p in self.protocol_type:
                if p not in e.keys():
                    continue
                elif len(self.protocol_list[p]) == 0 and e[p] is not None and e[p] != 'None':
                    # add first protocol for this protocol type
                    self.protocol_list[p].append(e[p])
                elif e[p] not in self.protocol_list[p] and e[p] is not None and e[p] != 'None':
                    # add other protocols for this protocol type
                    self.protocol_list[p].append(e[p])
                    # and this protocol type doesn't have unique value
                    self.protocol_unique[p] = False
            i = 'Data_Processing'
            if i in e.keys():
                # include experiment parameters in data processing part
                if len(self.other_item_list[i]) == 0 and e[i] is not None and e[i] != 'None':
                    self.other_item_list[i].append(e[i])
                elif e[i] not in self.other_item_list[i] and e[i] is not None and e[i] != 'None':
                    self.other_item_list[i].append(e[i])
                    self.other_item_unique[i] = False

        for e in self.experiment_settings:
            if e is None:
                continue
            i = 'Reference_Genome'
            if i in e.keys():
                if len(self.other_item_list[i]) == 0 and e[i] is not None and e[i] != 'None':
                    self.other_item_list[i].append(e[i])
                elif e[i] not in self.other_item_list[i] and e[i] is not None and e[i] != 'None':
                    self.other_item_list[i].append(e[i])
                    self.other_item_unique[i] = False

    # export data of all the samples in one study
    def export_sample(self):
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'SAMPLES', Style.SECTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1,
                              ('# This section lists and describes each of the biological Samples under investgation,'
                               'as well as any protocols that are specific to individual Samples.'), Style.INSTRUCTION)
        # write some blank cells to make words of last cell showing expanded.
        for col in range(2, 13):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1,
                              '# Additional "processed data file" or "raw file" columns may be included.',
                              Style.INSTRUCTION)
        # write some blank cells to make words of last cell showing expanded.
        for col in range(2, 13):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.header_row_idx = self.current_row_idx + 1
        self.current_row_idx += 1

        # write SAMPLES
        # experiment by experiment
        for e in self.experiments:
            if e is None or 'RID' not in e.keys():
                continue
            for r in self.replicates:
                if r is None or 'Experiment_RID' not in r.keys() or 'RID' not in r.keys():
                    continue
                elif r['Experiment_RID'] == e['RID']:
                    local_col_idx = 1
                    self.current_row_idx += 1
                    # construct sample_name using internal_ID,Biological_Replicate_Number and Technical_Replicate_Number
                    sample_name = e.get('Internal_ID', '') + '_' + str(
                        r.get('Biological_Replicate_Number', '')) + '_' + str(
                        r.get('Technical_Replicate_Number', ''))
                    sample_title = e.get('Name', '') + '_' + str(r.get('Biological_Replicate_Number', '')) + '_' + str(
                        r.get('Technical_Replicate_Number', ''))
                    sample_organism = e.get('Species', '')
                    sample_molecule = e.get('Molecule_Type', '')
                    sample_description = r.get('Notes', '')

                    self.excel.write_cell(self.header_row_idx, local_col_idx, 'Sample name', Style.HEADER)
                    self.excel.write_cell(self.current_row_idx, local_col_idx, sample_name)
                    local_col_idx += 1
                    self.excel.write_cell(self.header_row_idx, local_col_idx, 'title', Style.HEADER)
                    self.excel.write_cell(self.current_row_idx, local_col_idx, sample_title)
                    local_col_idx += 1

                    # the characteristic should show as one column in sample section, if samples has value it
                    characteristic_list = ['Phenotype', 'Stage_ID', 'Stage_Detail', 'Genotype', 'Strain', 'Wild_Type',
                        'Cell_Line','Passage','Assay_Type','Sex']
                    # foreign_item_list = ['Tissue', 'Source', 'Specimen_Cell_Type', 'Cell_Type']
                    characteristic_exist = []
                    for s in self.specimen:
                        for c in characteristic_list:
                            characteristic = s.get(c, '')
                            if characteristic is not None and len(characteristic) > 0 and c not in characteristic_exist:
                                # get a list of existing characteristics
                                characteristic_exist.append(c)
                            else:
                                continue

                    # check specimen tissue , if exist, add source
                    #/Src:=left(Tissue)=(Vocabulary:Anatomy:ID)
                    if self.source_type:
                        for f in self.source_type:
                            if f is None:
                                continue
                            else:
                                characteristic_exist.append('Source')
                                break
                    if self.cell_type:
                        for f in self.cell_type:
                            if f is None:
                                continue
                            else:
                                characteristic_exist.append('Cell_Type')
                                break
                    if self.specimen_allele:
                        for f in self.specimen_allele:
                            if f is None:
                                continue
                            else:
                                characteristic_exist.append('Allele')
                                break

                    # consolidated stage column
                    if 'Stage_ID' in characteristic_exist and 'Stage_Detail' in characteristic_exist:
                        characteristic_exist.remove('Stage_ID')
                        characteristic_exist.remove('Stage_Detail')
                        characteristic_exist.append('Stage')
                    elif 'Stage_ID' in characteristic_exist:
                        characteristic_exist.remove('Stage_ID')
                        characteristic_exist.append('Stage')
                    elif 'Stage_Detail' in characteristic_exist:
                        characteristic_exist.remove('Stage_Detail')
                        characteristic_exist.append('Stage')

                    # handling source name
                    for s in self.specimen:
                        if s is None or 'REPLICATE_RID' not in s.keys():
                            continue
                        elif s['RID'] == r['Specimen_RID']:
                            # handling source name
                            for c in characteristic_exist:
                                if c == "Source":
                                    source_type = ''
                                    for p in self.tissue:
                                        for q in self.source_type:
                                            if p is None or q is None:
                                                continue
                                            elif s['RID'] == p['Specimen_RID'] and p['Tissue'] == q['ID']:
                                                source_type = source_type + q['Name'] + ','
                                            else:
                                                continue
                                    characteristic = source_type[:-1] if len(source_type) > 0 else ''
                                    self.excel.write_cell(self.header_row_idx, local_col_idx, 'source name', Style.HEADER)
                                    self.excel.write_cell(self.current_row_idx, local_col_idx, characteristic)
                                    local_col_idx += 1

                    # handling organism
                    self.excel.write_cell(self.header_row_idx, local_col_idx, 'organism', Style.HEADER)
                    self.excel.write_cell(self.current_row_idx, local_col_idx, sample_organism)
                    local_col_idx += 1


                    for p in self.protocol_type:
                        # if different contents for protocol_type, need to add column for this protocol_type
                        # then different samples may has different content to write
                        if not self.protocol_unique[p]:
                            self.excel.write_cell(self.header_row_idx, local_col_idx, (p.replace('_', ' ').lower()),
                                                  Style.HEADER)
                            self.excel.write_cell(self.current_row_idx, local_col_idx, e[p])
                            local_col_idx += 1


                    for s in self.specimen:
                        if s is None or 'REPLICATE_RID' not in s.keys():
                            continue
                        elif s['RID'] == r['Specimen_RID']:
                            for c in characteristic_exist:
                                # handling characteristic cell_type
                                if c == "Cell_Type":
                                    cell_type = ''
                                    for p in self.specimen_cell_type:
                                        for q in self.cell_type:
                                            if p is None or q is None:
                                                continue
                                            elif s['RID'] == p['Specimen'] and p['Cell_Type'] == q['Name']:
                                                cell_type = cell_type + q['Name'] + ','
                                            else:
                                                continue
                                    characteristic = cell_type[:-1] if len(cell_type) > 0 else ''
                                    self.excel.write_cell(self.header_row_idx, local_col_idx,
                                                          'characteristics: Cell Type',
                                                          Style.HEADER)
                                    self.excel.write_cell(self.current_row_idx, local_col_idx, characteristic)
                                    local_col_idx += 1
                                # handling characteristic stage
                                elif c == "Stage":
                                    get_stage = 0
                                    for v in self.stage:
                                        if v is None:
                                            continue
                                        elif v['ID'] == s['Stage_ID']:
                                            characteristic = v['Name']
                                            self.excel.write_cell(self.header_row_idx, local_col_idx, 'characteristics: Stage', Style.HEADER)
                                            self.excel.write_cell(self.current_row_idx, local_col_idx, characteristic)
                                            local_col_idx += 1
                                            get_stage = 1
                                        else:
                                            continue
                                    # using Stage_Detail if no Stage_ID
                                    if get_stage == 0:
                                        characteristic = s.get('Stage_Detail', '')
                                        self.excel.write_cell(self.header_row_idx, local_col_idx, 'characteristics: Age', Style.HEADER)
                                        self.excel.write_cell(self.current_row_idx, local_col_idx, characteristic)
                                        local_col_idx += 1
                                
                                # Handling Specimen Allele Information
                                elif c == 'Allele':
                                    allele_info = []
                                    for sa in self.specimen_allele:
                                        if sa is None:
                                            continue
                                        elif s['RID'] == sa['Specimen_RID']:
                                                for a in self.allele:
                                                    if a is None:
                                                        continue
                                                    elif sa['Allele_RID'] == a['RID']:
                                                        if a.get('Allele_Type') is not None:
                                                            allele_info.append( a['Name'] + ' (' + a['Allele_Type'] + ')' )
                                                        else:
                                                            allele_info.append( a['Name'] )
                                    
                                    characteristic = ', '.join( allele_info ) if len(allele_info ) > 0 else ''
                                    self.excel.write_cell(self.header_row_idx, local_col_idx, 'characteristics: Allele', Style.HEADER )
                                    self.excel.write_cell(self.current_row_idx, local_col_idx, characteristic )
                                    local_col_idx += 1
                                    
                                else:
                                    characteristic = s.get(c, '')
                                    self.excel.write_cell(self.header_row_idx, local_col_idx, 'characteristics: '+c, Style.HEADER)
                                    self.excel.write_cell(self.current_row_idx, local_col_idx, characteristic)
                                    local_col_idx += 1
                        else:
                            continue


                    self.excel.write_cell(self.header_row_idx, local_col_idx, 'molecule', Style.HEADER)
                    self.excel.write_cell(self.current_row_idx, local_col_idx, sample_molecule)
                    local_col_idx += 1

                    self.excel.write_cell(self.header_row_idx, local_col_idx, 'description', Style.HEADER)
                    self.excel.write_cell(self.current_row_idx, local_col_idx, sample_description)
                    local_col_idx += 1

                    
                    
                    # processed files in template are write to supplementary file
                    # 1. set flagWhitelist to False then all the files will pass
                    # 2. set flagWhitelist to True and fileEndingList to [] then no file will pass
                    # 3. set flagWhitelist to True and fileEndingList to [xxx,yyy] then only files ending with xxx,yyy will pass
                    flagWhitelist = True
                    fileTypeWhiteList = [ ".bam",".fastq",".fastq.gz" ] 
                    
                    for f in self.files:
                        validFile = False
                        if f is None or 'Replicate_RID' not in f.keys():
                            continue
                        elif flagWhitelist and fileTypeWhiteList:
                            for extension in fileTypeWhiteList:
                                if f.get('File_Name') is not None and f.get('File_Name').lower().endswith(extension):
                                    validFile = True
                        else:
                            # if flagWhitelist == False then all the files are valid
                            validFile = True
                        
                        if validFile and f['Replicate_RID'] == r['RID']:
                            file_path = self.parameters.get("replicate_files_path_template", f.get("File_Name", ""))
                            self.excel.write_cell(self.header_row_idx, local_col_idx, 'raw file', Style.HEADER)
                            self.excel.write_cell(self.current_row_idx, local_col_idx, file_path.format(**f))
                            local_col_idx += 1
                        else:
                            continue
                else:
                    continue

    # export data for protocol section
    def export_protocol(self):
        self.current_row_idx = self.current_row_idx + 2
        self.excel.write_cell(self.current_row_idx, 1, 'PROTOCOLS', Style.SECTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1,
                              ('# Any of the protocols below which are applicable to only a subset of '
                               'Samples should be included as additional columns of the SAMPLES section instead.'),
                              Style.INSTRUCTION)
        # write some blank cells to make words of last cell showing expanded.
        for col in range(2, 13):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1

        self.excel.write_cell(self.current_row_idx, 1, 'growth protocol', Style.FIELD)
        if self.protocol_unique['Growth_Protocol']:
            if len(self.protocol_list['Growth_Protocol']) == 1:
                self.excel.write_cell(self.current_row_idx, 2, self.protocol_list['Growth_Protocol'][0])
            else:
                self.excel.write_cell(self.current_row_idx, 2, '')
        else:
            self.excel.write_cell(self.current_row_idx, 2, '')
        self.current_row_idx += 1

        self.excel.write_cell(self.current_row_idx, 1, 'treatment protocol', Style.FIELD)
        if self.protocol_unique['Treatment_Protocol']:
            if len(self.protocol_list['Treatment_Protocol']) == 1:
                self.excel.write_cell(self.current_row_idx, 2, self.protocol_list['Treatment_Protocol'][0])
            else:
                self.excel.write_cell(self.current_row_idx, 2, '')
        else:
            self.excel.write_cell(self.current_row_idx, 2, '')
        self.current_row_idx += 1

        self.excel.write_cell(self.current_row_idx, 1, 'extract protocol', Style.FIELD)
        if self.protocol_unique['Extract_Protocol']:
            if len(self.protocol_list['Extract_Protocol']) == 1:
                self.excel.write_cell(self.current_row_idx, 2, self.protocol_list['Extract_Protocol'][0])
            else:
                self.excel.write_cell(self.current_row_idx, 2, '')
        else:
            self.excel.write_cell(self.current_row_idx, 2, '')
        self.current_row_idx += 1

        self.excel.write_cell(self.current_row_idx, 1, 'library construction protocol', Style.FIELD)
        if self.protocol_unique['Construction_Protocol']:
            if len(self.protocol_list['Construction_Protocol']) == 1:
                self.excel.write_cell(self.current_row_idx, 2, self.protocol_list['Construction_Protocol'][0])
            else:
                self.excel.write_cell(self.current_row_idx, 2, '')
        else:
            self.excel.write_cell(self.current_row_idx, 2, '')
        self.current_row_idx += 1

        for p in self.protocol_type:
            if p not in ['Growth_Protocol', 'Treatment_Protocol', 'Extract_Protocol', 'Construction_Protocol']:
                # only write protocol_types of unique content for all the samples in the same study
                # as protocl_types of non-unique content are already exported in sample section
                if self.protocol_unique[p] and len(self.protocol_list[p]) == 1:
                    self.excel.write_cell(self.current_row_idx, 1, (p.replace('_', ' ').lower()), Style.FIELD)
                    self.excel.write_cell(self.current_row_idx, 2, self.protocol_list[p][0])
                    self.current_row_idx += 1

        self.excel.write_cell(self.current_row_idx, 1, 'library strategy', Style.FIELD)

        library_type_list = []
        for es in self.experiment_settings:
            if es is None or 'Experiment_RID' not in es.keys():
                continue
            else:
                if es['Library_Type'] not in library_type_list:
                    library_type_list.append(es['Library_Type'])

        self.excel.write_cell(self.current_row_idx, 2, ','.join(library_type_list))

    # export data processing pipline information
    def export_data_processing(self):
        # DATA PROCESSING PIPELINE
        self.current_row_idx = self.current_row_idx + 2

        self.excel.write_cell(self.current_row_idx, 1, 'DATA PROCESSING PIPELINE', Style.SECTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1,
                              ('# Data processing steps include base-calling, alignment, filtering,'
                               'peak-calling, generation of normalized abundance measurements etc'), Style.INSTRUCTION)
        for col in range(2, 13):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1,
                              ('# For each step provide a description, as well as software name,'
                               'version, parameters, if applicable.'), Style.INSTRUCTION)
        for col in range(2, 13):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, '# Include additional steps, as necessary.', Style.INSTRUCTION)
        for col in range(2, 13):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1


        if len(self.other_item_list['Data_Processing']) == 1:
            processing_list = self.other_item_list['Data_Processing'][0].split('\n')
        else:
            processing_list = []

        # include experiment setting parameters into data processing section
        other_procession_set = {}
        for es in self.experiment_settings:
            if es is None or 'Experiment_RID' not in es.keys():
                continue
            else:
                if other_procession_set:
                    for k, v in es.items():
                        if k in other_procession_set.keys() and v != other_procession_set[k]:
                            del other_procession_set[k]
                else:
                    other_procession_set = es.copy()

        for step_row in processing_list:
            self.excel.write_cell(self.current_row_idx, 1, 'data processing step', Style.FIELD)
            self.excel.write_cell(self.current_row_idx, 2, step_row)
            self.current_row_idx += 1

        for k, v in other_procession_set.items():
            if v and k not in ['RCB', 'RCT', 'RID', 'RMB', 'RMT'] and len(v) > 0:
                if k in ['Library','Visualization','Alignment','Sequencing_Method','Quantification'] and 'detail' in v:
                    continue
                self.excel.write_cell(self.current_row_idx, 1, 'data processing step', Style.FIELD)
                self.excel.write_cell(self.current_row_idx, 2, k + ' : ' + v)
                self.current_row_idx += 1

        if self.other_item_unique['Reference_Genome']:
            if len(self.other_item_list['Reference_Genome']) == 1:
                self.excel.write_cell(self.current_row_idx, 1, 'genome build', Style.FIELD)
                self.excel.write_cell(self.current_row_idx, 2, self.other_item_list['Reference_Genome'][0])
            else:
                self.excel.write_cell(self.current_row_idx, 1, 'genome build', Style.FIELD)
                self.excel.write_cell(self.current_row_idx, 2, ' '.join(self.other_item_list['Reference_Genome']))
                logging.debug('Multi-genome')

        self.current_row_idx += 1

        self.excel.write_cell(self.current_row_idx, 1, 'processed data files format and content', Style.FIELD)
        self.excel.write_cell(self.current_row_idx, 2, 'processed data file format')
        self.current_row_idx += 1

    
    # export processed datafile names of this study
    def export_processed_datafiles(self):
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1,
                              ('# For each file listed in the "processed data file" columns of '
                               'the SAMPLES section, provide additional information below.'), Style.INSTRUCTION)
        for col in range(2, 13):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'PROCESSED DATA FILES', Style.SECTION)
        self.header_row_idx = self.current_row_idx + 1
        self.current_row_idx += 2


        # 1. set flagWhitelist to False then all the files will pass
        # 2. set flagWhitelist to True and fileEndingList to [] then no file will pass
        # 3. set flagWhitelist to True and fileEndingList to [xxx,yyy] then only files ending with xxx,yyy will pass
        flagWhitelist = True
        # Adding .bw to the Study_File whitelist
        fileTypeWhiteList = [".csv",".csv.gz",".txt.gz",".txt",".tsv",".tsv.gz",".xls",".xlsx",".mtx",".bw"]

        for pf in self.study_files:
            validFile = False
            if pf is None:
                continue
            elif flagWhitelist:
                for str in fileTypeWhiteList:
                    if pf.get('File_Name') is not None and pf.get('File_Name').lower().endswith(str):
                        validFile = True
            else:
                # if flagWhitelist == False then all the files are valid
                validFile = True
            if validFile:
                local_col_idx = 1
                file_path = self.parameters.get("study_files_path_template", pf.get("File_Name", ""))
                file_path = file_path.format(**pf)
                self.excel.write_cell(self.header_row_idx, local_col_idx, 'file name', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, local_col_idx, file_path)
                local_col_idx += 1
                self.excel.write_cell(self.header_row_idx, local_col_idx, 'file type', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, local_col_idx, pf.get('File_Type', ''))
                local_col_idx += 1
                self.excel.write_cell(self.header_row_idx, local_col_idx, 'file checksum', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, local_col_idx, pf.get('MD5', ''))
                self.current_row_idx += 1


    # export raw datafile names of this study
    def export_raw_datafiles(self):
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, ('# For each file listed in the "raw file" columns of'
                                                        'the SAMPLES section, provide additional information below.'),
                              Style.INSTRUCTION)
        for col in range(2, 13):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'RAW FILES', Style.SECTION)
        self.header_row_idx = self.current_row_idx + 1
        self.current_row_idx += 2


        # Group the File Names by Replicate RID's
        replicateFileDict = {}    
        for pf in self.files:
            if pf.get('Replicate_RID') is None or pf.get('File_Name') is None:
                continue
            elif pf['Replicate_RID'] in replicateFileDict:
                replicateFileDict[ pf['Replicate_RID'] ].append( pf['File_Name'] )
            else:
                replicateFileDict[ pf['Replicate_RID'] ] = [ pf['File_Name'] ]


        derivedPairedEnds = {}
        # fileTypeWhiteList = {} # https://github.com/informatics-isi-edu/rbk-project/issues/652
        patternR1 = ".*(\.|_)R1(\.|_)[0-9]*(\.|_)*(FASTQ|FQ).GZ"
        patternR2 = ".*(\.|_)R2(\.|_)[0-9]*(\.|_)*(FASTQ|FQ).GZ"
        for replicate_rid, fileList in replicateFileDict.items():
            ## Derive the paired end values using replicate level files.
            ## https://github.com/informatics-isi-edu/rbk-project/issues/615
            if ( any( re.match( patternR1, fileName.upper() ) for fileName in fileList ) 
                and any( re.match( patternR2, fileName.upper() ) for fileName in fileList ) ):
                derivedPairedEnds[ replicate_rid ] = 'paired-end'
            elif any( re.match( patternR1, fileName.upper() ) for fileName in fileList ):
                derivedPairedEnds[ replicate_rid ] = 'single'
            else:
                derivedPairedEnds[ replicate_rid ] = None

            ## GEO prefer bams/sams, but will accept fastq’s if that is all that is present. 
            ## Add logic to set priority of seq files to export
            ## https://github.com/informatics-isi-edu/rbk-project/issues/652
            ##
            ## THIS CHANGE IS PUT ON HOLD!! 
            ## until we can find a way to filter the files in the export BDBags as well to keep the data consistent.
            ##############################################################################################################
            
            # if ( any( re.match( '.*\.(bam|sam|bai)', fileName.lower() ) for fileName in fileList ) ):
            #     fileTypeWhiteList[ replicate_rid ] = [ ".bam" ]
            # else:
            #     fileTypeWhiteList[ replicate_rid ] = [ ".fastq", ".fastq.gz" ]


        # 1. set flagWhitelist to False then all the files will pass
        # 2. set flagWhitelist to True and fileEndingList to [] then no file will pass
        # 3. set flagWhitelist to True and fileEndingList to [xxx,yyy] then only files ending with xxx,yyy will pass
        flagWhitelist = True
        # Removing .bai from whitelist, GEO will ignore it as per Gervaise.
        fileTypeWhiteList = [ ".bam",".fastq",".fastq.gz" ] 
        for pf in self.files:
            validFile = False
            if pf is None or 'Experiment_RID' not in pf.keys():
                continue
            elif flagWhitelist and fileTypeWhiteList:
                # for str in fileTypeWhiteList[ pf.get('Replicate_RID' ) ]: # https://github.com/informatics-isi-edu/rbk-project/issues/652
                for str in fileTypeWhiteList:
                    if pf.get('File_Name') is not None and pf.get('File_Name').lower().endswith(str):
                        validFile = True
            else:
                # if flagWhitelist == False then all the files are valid
                validFile = True
            if validFile:
                current_col_idx = 1
                file_path = self.parameters.get("replicate_files_path_template", pf.get("File_Name", ""))
                self.excel.write_cell(self.header_row_idx, current_col_idx, 'file name', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, current_col_idx, file_path.format(**pf))
                current_col_idx += 1
                self.excel.write_cell(self.header_row_idx, current_col_idx, 'file type', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, current_col_idx, pf.get('File_Type', ''))
                current_col_idx += 1
                self.excel.write_cell(self.header_row_idx, current_col_idx, 'file checksum', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, current_col_idx, pf.get('MD5', ''))
                current_col_idx += 1
                # Experiment_Settings.Protocol_Reference
                for es in self.experiment_settings:
                    if es is None or 'Experiment_RID' not in es.keys():
                        continue
                    elif es['Experiment_RID'] == pf['Experiment_RID']:
                        self.excel.write_cell(self.header_row_idx, current_col_idx, 'instrument model', Style.HEADER)
                        self.excel.write_cell(self.current_row_idx, current_col_idx, es.get('Sequencing_Platform', ''))
                        current_col_idx += 1
                        
                        
                        ## Remove Read_length from the excel ( https://github.com/informatics-isi-edu/rbk-project/issues/667 )
                        ## If we do add Read Length GEO team will either ignore it (raw file section) or print it if you put it in the data processing.
                        # self.excel.write_cell(self.header_row_idx, current_col_idx, 'read length', Style.HEADER)
                        # self.excel.write_cell(self.current_row_idx, current_col_idx, es.get('Read Length', ''))
                        # current_col_idx += 1
                        
                        ## Modify the paired end logic to fetch derived values.
                        ## https://github.com/informatics-isi-edu/rbk-project/issues/615
                        ## OLD CODE
                        # if es.get('Paired_End') is not None and 'pair' in es.get('Paired_End').lower():
                        #     single_or_paired = 'paired-end'
                        # else:
                        #     single_or_paired = 'single'
                        
                        ## NEW CODE
                        # If the paired end value could not be derived then switch to the value in Experiment Settings
                        if ( derivedPairedEnds is None
                            or pf.get('Replicate_RID') is None
                            or pf['Replicate_RID'] not in derivedPairedEnds
                            or derivedPairedEnds [ pf['Replicate_RID'] ] is None ):
                            ## OLD CODE
                            if es.get('Paired_End') is not None and 'pair' in es.get('Paired_End').lower():
                                single_or_paired = 'paired-end'
                            else:
                                single_or_paired = 'single'
                        else:
                            single_or_paired = derivedPairedEnds[ pf['Replicate_RID'] ]                        
                        ## MODIFICATION END

                        self.excel.write_cell(self.header_row_idx, current_col_idx, 'single or paired-end',
                                              Style.HEADER)
                        self.excel.write_cell(self.current_row_idx, current_col_idx, single_or_paired)
                        current_col_idx += 1

                self.current_row_idx += 1

    def export_paired_end(self):
        # PAIRED-END EXPERIMENTS
        # todo need check what's paired-end data look like
        # currently, we haven't seen paired-end data
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, (
            '# For paired-end experiments, list the 2 associated raw files, and provide average insert size and standard deviation,'
            'if known. For SOLiD experiments, list the 4 file names (include "file name 3" and "file name 4" columns).'),
                              Style.INSTRUCTION)
        for col in range(2, 13):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'PAIRED-END EXPERIMENTS', Style.SECTION)
        self.header_row_idx = self.current_row_idx + 1
        self.current_row_idx += 1
        current_col_idx = 1
        self.excel.write_cell(self.header_row_idx, current_col_idx, 'file name 1', Style.HEADER)
        current_col_idx += 1
        self.excel.write_cell(self.header_row_idx, current_col_idx, 'file name 2', Style.HEADER)
        current_col_idx += 1
        self.excel.write_cell(self.header_row_idx, current_col_idx, 'average insert size', Style.HEADER)
        current_col_idx += 1
        self.excel.write_cell(self.header_row_idx, current_col_idx, 'standard deviation', Style.HEADER)

    def export_finish(self):
        self.excel.save_xlsx()



import sys
from deriva.transfer import DerivaUploadCLI, GenericUploader

DESC = "Deriva Data Upload Utility - CLI"
INFO = "For more information see: https://github.com/informatics-isi-edu/deriva-py"


def main():
    cli = DerivaUploadCLI(GenericUploader, DESC, INFO)
    return cli.main()


if __name__ == '__main__':
    sys.exit(main())



import os
import json
import uuid
import logging
import requests
from bdbag import bdbag_ro as ro
from deriva.core import format_exception
from deriva.core.utils.hash_utils import decodeBase64toHex
from deriva.core.utils.mime_utils import parse_content_disposition
from deriva.transfer.download.processors.query.base_query_processor import BaseQueryProcessor, LOCAL_PATH_KEY
from deriva.transfer.download import DerivaDownloadError, DerivaDownloadConfigurationError


class BagFetchQueryProcessor(BaseQueryProcessor):
    def __init__(self, envars=None, **kwargs):
        super(BagFetchQueryProcessor, self).__init__(envars, **kwargs)
        self.content_type = "application/x-json-stream"
        filename = ''.join(['fetch-manifest_', str(uuid.uuid4()), ".json"])
        self.output_relpath, self.output_abspath = self.create_paths(self.base_path, filename=filename)
        self.ro_file_provenance = False

    def process(self):
        super(BagFetchQueryProcessor, self).process()
        rfm_relpath, rfm_abspath = self.createRemoteFileManifest()
        if rfm_relpath and rfm_abspath:
            self.outputs.update({rfm_relpath: {LOCAL_PATH_KEY: rfm_abspath}} if not self.is_bag else {})
        return self.outputs

    def createRemoteFileManifest(self):
        logging.info("Creating remote file manifest from results of query: %s" % self.query)
        input_manifest = self.output_abspath
        remote_file_manifest = self.kwargs.get("remote_file_manifest")

        if not os.path.isfile(input_manifest):
            return None, None

        with open(input_manifest, "r") as in_file, open(remote_file_manifest, "a") as remote_file:
            for line in in_file:
                # get the required bdbag remote file manifest vars from each line of the json-stream input file
                entry = json.loads(line)
                entry = self.createManifestEntry(entry)
                if not entry:
                    continue
                remote_file.write(json.dumps(entry) + "\n")
                if self.ro_manifest:
                    ro.add_file_metadata(self.ro_manifest,
                                         source_url=entry["url"],
                                         media_type=entry.get("content_type"),
                                         bundled_as=ro.make_bundled_as(
                                             folder=os.path.dirname(entry["filename"]),
                                             filename=os.path.basename(entry["filename"])))
        os.remove(input_manifest)
        return os.path.relpath(remote_file_manifest, self.base_path), os.path.abspath(remote_file_manifest)

    def createManifestEntry(self, entry):
        manifest_entry = dict()
        url = entry.get("url")
        if not url:
            logging.warning(
                "Skipping a record due to missing required attribute \"url\" in fetch manifest entry %s" %
                json.dumps(entry))
            return

        ext_url = self.getExternalUrl(url)
        length = entry.get("length")
        md5 = entry.get("md5")
        sha256 = entry.get("sha256")
        filename = entry.get("filename") if not self.output_filename else self.output_filename
        content_type = entry.get("content_type")
        content_disposition = None
        # if any required fields are missing from the query result, attempt to get them from the remote server by
        # issuing a HEAD request against the supplied URL
        if not (length and (md5 or sha256)):
            try:
                headers = self.headForHeaders(url, raise_for_status=True)
            except requests.HTTPError as e:
                raise DerivaDownloadError("Exception during HEAD request: %s" % format_exception(e))
            length = headers.get("Content-Length")
            content_type = headers.get("Content-Type")
            content_disposition = headers.get("Content-Disposition")
            if not md5:
                md5 = headers.get("Content-MD5")
                if md5:
                    md5 = decodeBase64toHex(md5)
            if not sha256:
                sha256 = headers.get("Content-SHA256")
                if sha256:
                    sha256 = decodeBase64toHex(sha256)
        # if content length or both hash values are missing, it is a fatal error
        if not length:
            raise DerivaDownloadError("Could not determine Content-Length for %s" % ext_url)
        if not (md5 or sha256):
            raise DerivaDownloadError("Could not locate an MD5 or SHA256 hash for %s" % ext_url)
        # if a local filename is not provided, try to construct one using content_disposition, if available
        if not filename:
            filename = os.path.basename(url).split(":")[0] if not content_disposition else \
                parse_content_disposition(content_disposition)
        env = self.envars.copy()
        env.update(entry)
        output_path, _ = self.create_paths(self.base_path,
                                           sub_path=self.sub_path,
                                           filename=filename,
                                           is_bag=self.is_bag,
                                           envars=env)
        manifest_entry['url'] = ext_url
        manifest_entry['length'] = int(length)
        manifest_entry['filename'] = output_path
        if md5:
            manifest_entry['md5'] = md5
        if sha256:
            manifest_entry['sha256'] = sha256
        if content_type:
            manifest_entry["content_type"] = content_type
        return manifest_entry




import os
import io
import logging
import datetime
import codecs
import csv
import json

from . import urlquote, datapath, DEFAULT_HEADERS, DEFAULT_CHUNK_SIZE, Megabyte, Kilobyte, get_transfer_summary, IS_PY2
from .deriva_binding import DerivaBinding
from . import ermrest_model


class ErmrestCatalogMutationError(Exception):
    pass


_clone_state_url = "tag:isrd.isi.edu,2018:clone-status"

DEFAULT_PAGE_SIZE = 100000


class ErmrestCatalog(DerivaBinding):
    """Persistent handle for an ERMrest catalog.

       Provides basic REST client for HTTP methods on arbitrary
       paths. Caller has to understand ERMrest APIs and compose
       appropriate paths, headers, and/or content.

       Additional utility methods provided for accessing catalog metadata.
    """
    table_schemas = dict()

    def __init__(self, scheme, server, catalog_id, credentials=None, caching=True, session_config=None):
        """Create ERMrest catalog binding.

           Arguments:
             scheme: 'http' or 'https'
             server: server FQDN string
             catalog_id: e.g. '1'
             credentials: credential secrets, e.g. cookie
             caching: whether to retain a GET response cache

           Deriva Client Context: You MAY mutate self.dcctx to
           customize the context for this service endpoint prior to
           invoking web requests.  E.g.:

             self.dcctx['cid'] = 'my application name'

           You MAY also supply custom per-request context by passing a
           headers dict to web request methods, e.g. 

             self.get(..., headers={'deriva-client-context': {'action': 'myapp/function1'}})

           This custom header will be merged as override values with
           the default context in self.dcctx in order to form the
           complete context for the request.
        """
        super(ErmrestCatalog, self).__init__(scheme, server, credentials, caching, session_config)
        self._server_uri = "%s/ermrest/catalog/%s" % (
            self._server_uri,
            catalog_id
        )
        self._scheme, self._server, self._catalog_id, self._credentials, self._caching, self._session_config = \
            scheme, server, catalog_id, credentials, caching, session_config

    @property
    def catalog_id(self):
        return self._catalog_id

    def exists(self):
        """Simple boolean test for catalog existence.

        :return: True if exists, False if not (404), otherwise raises exception
        """
        try:
            self.get('/')
            return True
        except HTTPError as e:
            if e.response.status_code == 404:
                return False
            else:
                raise

    def latest_snapshot(self):
        """Gets a handle to this catalog's latest snapshot.
        """
        r = self.get('/')
        r.raise_for_status()
        return ErmrestSnapshot(self._scheme, self._server, self._catalog_id, r.json()['snaptime'],
                               self._credentials, self._caching, self._session_config)

    def getCatalogModel(self):
        return ermrest_model.Model.fromcatalog(self)

    def getCatalogSchema(self):
        path = '/schema'
        r = self.get(path)
        r.raise_for_status()
        return r.json()

    def getPathBuilder(self):
        """Returns the 'path builder' interface for this catalog."""
        return datapath.from_catalog(self)

    def getTableSchema(self, fq_table_name):
        # first try to get from cache(s)
        s, t = self.splitQualifiedCatalogName(fq_table_name)
        cat = self.getCatalogSchema()
        schema = cat['schemas'][s]['tables'][t] if cat else None
        if schema:
            return schema
        schema = self.table_schemas.get(fq_table_name)
        if schema:
            return schema

        path = '/schema/%s/table/%s' % (s, t)
        r = self.get(path)
        resp = r.json()
        self.table_schemas[fq_table_name] = resp
        r.raise_for_status()

        return resp

    def getTableColumns(self, fq_table_name):
        columns = set()
        schema = self.getTableSchema(fq_table_name)
        for column in schema['column_definitions']:
            columns.add(column['name'])

        return columns

    def validateRowColumns(self, row, fq_tableName):
        columns = self.getTableColumns(fq_tableName)
        return set(row.keys()) - columns

    def getDefaultColumns(self, row, table, exclude=None, quote_url=True):
        columns = self.getTableColumns(table)
        if isinstance(exclude, list):
            for col in exclude:
                columns.remove(col)

        defaults = []
        supplied_columns = row.keys()
        for col in columns:
            if col not in supplied_columns:
                defaults.append(urlquote(col, safe='') if quote_url else col)

        return defaults

    @staticmethod
    def splitQualifiedCatalogName(name):
        entity = name.split(':')
        if len(entity) != 2:
            logging.debug("Unable to tokenize %s into a fully qualified <schema:table> name." % name)
            return None
        return entity[0], entity[1]

    def getAsFile(self,
                  path,
                  destfilename,
                  headers=DEFAULT_HEADERS,
                  callback=None,
                  delete_if_empty=False,
                  paged=False,
                  page_size=DEFAULT_PAGE_SIZE):
        """
           Retrieve catalog data streamed to destination file.
           Caller is responsible to clean up file even on error, when the file may or may not be exist.
           If "delete_if_empty" is True, the file will be inspected for "empty" content. In the case of
           json/json-stream content, the presence of a single empty JSON object will be tested for. In the case of
           CSV content, the file will be parsed with CSV reader to determine that only a single header line and no row
           data is present.
        """
        self.check_path(path)

        # Only entity API supported with paged mode at this time, otherwise fallback. We fallback rather than raise an
        # exception in the case that the caller might be trying to perform an opportunistic paged request without
        # knowing a priori if paged support for the given query is available.
        page_size = page_size if page_size > 0 else DEFAULT_PAGE_SIZE
        if not path.startswith("/entity") and paged:
            logging.debug("Paged data retrieval only supported for entity API queries.")
            paged = False

        # Only "application/x-json-stream" or "text/csv" supported with paged mode at this time, otherwise fallback.
        accept = headers.get("accept")
        if accept not in ("application/x-json-stream", "text/csv"):
            logging.debug("Paged data retrieval not supported for content type: %s" % accept)
            paged = False

        headers = headers.copy()

        destfile = open(destfilename, 'w+b')

        try:
            total = 0
            start = datetime.datetime.now()
            if not paged:
                with self._session.get(self._server_uri + path, headers=headers, stream=True) as r:
                    self._response_raise_for_status(r)
                    content_type = r.headers.get("Content-Type")
                    logging.debug("Transferring file %s to %s" % (self._server_uri + path, destfilename))
                    for buf in r.iter_content(chunk_size=DEFAULT_CHUNK_SIZE):
                        destfile.write(buf)
                        total += len(buf)
                        if callback:
                            if not callback(progress="Downloading: %.2f MB transferred" %
                                                     (float(total) / float(Megabyte))):
                                destfile.close()
                                return
                destfile.flush()
                os.fsync(destfile.fileno())
            else:
                first_page = True
                first_line = None
                last_record = None
                while True:
                    qualifier = "@sort(RID)%s?limit=%d" % \
                               (("@after(%s)" % urlquote(last_record)) if last_record is not None else "", page_size)
                    # 1. Try to get a page worth of data, back-off page size if query run time errors are encountered
                    with self._session.get(self._server_uri + path + qualifier, headers=headers) as r:
                        if r.status_code == 400 and "Query run time limit exceeded" in r.text:
                            r.close()
                            page_size //= 2
                            page_size = 1 if page_size < 1 else page_size
                            logging.warning("Query runtime exceeded while attempting to transfer rows from %s to file "
                                            "[%s]. The page size is being reduced to %s and the query will be retried."
                                            % (self._server_uri + path, destfilename, page_size))
                            continue

                        # 2. Write the page to disk and determine the last RID processed in order to get the next page
                        last_line = None
                        content_type = r.headers.get("Content-Type")
                        logging.debug("Transferring file %s to %s" % (self._server_uri + path, destfilename))
                        # CSV processing iterates over lines in the response, skipping the header line(s) in all but
                        # the first page, and captures the last line of each page to determine the last record processed
                        if content_type == "text/csv":
                            skip = 1
                            line_num = 0
                            if first_page:
                                lines = r.iter_lines(decode_unicode=True)
                                reader = csv.reader(lines)
                                first_line = next(reader)
                                skip = reader.line_num
                            for line in r.iter_lines():
                                if not first_page:
                                    line_num += 1
                                    if line_num <= skip:
                                        continue
                                tline = line + b"\n"
                                destfile.write(tline)
                                total += len(tline)
                                last_line = tline
                            if last_line and last_line != first_line:
                                reader = csv.DictReader([last_line.decode('utf-8')], first_line)
                                last_line = next(reader)
                            first_page = False
                        # JSON-Stream processing writes the entire buffer to the destination file. The last line is
                        # captured by reverse seeking in the buffer from right before the last b'\n' newline to the next
                        # newline or buf[0], then calling readline from the current position
                        elif content_type == "application/x-json-stream":
                            buf = r.content
                            if not buf:
                                break
                            destfile.write(buf)
                            total += len(buf)
                            b = io.BytesIO(buf)
                            b.seek(-2, os.SEEK_END)
                            while b.read(1) != b'\n':
                                b.seek(-2, os.SEEK_CUR)
                                if b.tell() == os.SEEK_SET:
                                    break
                            last_line = json.loads(b.readline().decode('utf-8'))

                        # 3. Save the last record RID and flush the destination file buffers to disk.
                        if not last_line:
                            break
                        destfile.flush()
                        os.fsync(destfile.fileno())
                        last_record = last_line['RID']
                        if callback:
                            if not callback(progress="Downloading: %.2f MB transferred" %
                                                     (float(total) / float(Megabyte))):
                                destfile.close()
                                return

            elapsed = datetime.datetime.now() - start
            summary = get_transfer_summary(total, elapsed)

            # perform automatic file deletion on detected "empty" content, if requested
            delete_file = True if total == 0 else False
            if delete_if_empty and total > 0:
                destfile.seek(0)
                if content_type == "application/json" or content_type == "application/x-json-stream":
                    buf = destfile.read(16)
                    if buf == b"[]\n" or buf == b"{}\n":
                        delete_file = True
                elif content_type == "text/csv":
                    reader = csv.reader(codecs.iterdecode(destfile, 'utf-8') if not IS_PY2 else destfile)
                    rowcount = 0
                    for row in reader:
                        rowcount += 1
                        if rowcount > 1:
                            break
                    if rowcount <= 1:
                        delete_file = True

            # automatically delete zero-length files or detected "empty" content
            if delete_file:
                destfile.close()
                os.remove(destfilename)
                destfile = None

            log_msg = "File [%s] transfer successful. %s %s" % \
                      (destfilename, summary,
                       "File was automatically deleted due to empty content." if delete_file else "")
            logging.info(log_msg)
            if callback:
                callback(summary=log_msg, file_path=destfilename)

        finally:
            if destfile:
                destfile.close()

    def delete(self, path, headers=DEFAULT_HEADERS, guard_response=None):
        """Perform DELETE request, returning response object.

           Arguments:
             path: the path within this bound catalog
             headers: headers to set in request
             guard_response: expected current resource state
               as previously seen response object.

           Uses guard_response to build appropriate 'if-match' header
           to assure change is only applied to expected state.

           Raises ConcurrentUpdate for 412 status.

        """
        if path == "/":
            raise DerivaPathError('See self.delete_ermrest_catalog() if you really want to destroy this catalog.')
        return DerivaBinding.delete(self, path, headers=headers, guard_response=guard_response)

    def delete_ermrest_catalog(self, really=False):
        """Perform DELETE request, destroying catalog on server.

           Arguments:
             really: delete when True, abort when False (default)

        """
        if really is True:
            return DerivaBinding.delete(self, '/')
        else:
            raise ValueError('Catalog deletion refused when really is %s.' % really)

    def clone_catalog(self,
                      dst_catalog=None,
                      copy_data=True,
                      copy_annotations=True,
                      copy_policy=True,
                      truncate_after=True,
                      exclude_schemas=None):
        """Clone this catalog's content into dest_catalog, creating a new catalog if needed.

        :param dst_catalog: Destination catalog or None to request creation of new destination (default).
        :param copy_data: Copy table contents when True (default).
        :param copy_annotations: Copy annotations when True (default).
        :param copy_policy: Copy access-control policies when True (default).
        :param truncate_after: Truncate destination history after cloning when True (default).
        :param exclude_schemas: A list of schema names to exclude from the cloning process.

        When dest_catalog is provided, attempt an idempotent clone,
        assuming content MAY be partially cloned already using the
        same parameters. This routine uses a table-level annotation
        "tag:isrd.isi.edu,2018:clone-state" to save progress markers
        which help it restart efficiently if interrupted.

        Cloning preserves source row RID values so that any RID-based
        foreign keys are still valid. It is not generally advisable to
        try to merge more than one source into the same clone, nor to
        clone on top of rows generated locally in the destination,
        since this could cause duplicate RID conflicts.

        Truncation after cloning avoids retaining incremental
        snapshots which contain partial clones.

        """
        src_model = self.getCatalogModel()

        if dst_catalog is None:
            # TODO: refactor with DerivaServer someday
            server = DerivaBinding(self._scheme, self._server, self._credentials, self._caching, self._session_config)
            dst_id = server.post("/ermrest/catalog").json()["id"]
            dst_catalog = ErmrestCatalog(self._scheme, self._server, dst_id, self._credentials, self._caching, self._session_config)

        # set top-level config right away and find fatal usage errors...
        if copy_policy:
            if not src_model.acls:
                raise ValueError("Use of copy_policy=True not possible when caller does not own source catalog.")
            dst_catalog.put('/acl', json=src_model.acls)

        if copy_annotations:
            dst_catalog.put('/annotation', json=src_model.annotations)

        # build up the model content we will copy to destination
        dst_model = dst_catalog.getCatalogModel()

        new_model = []
        new_columns = [] # ERMrest does not currently allow bulk column creation
        new_keys = [] # ERMrest does not currently allow bulk key creation
        clone_states = {}
        fkeys_deferred = {}
        exclude_schemas = [] if exclude_schemas is None else exclude_schemas

        def prune_parts(d):
            if not copy_annotations and 'annotations' in d:
                del d['annotations']
            if not copy_policy:
                if 'acls' in d:
                    del d['acls']
                if 'acl_bindings' in d:
                    del d['acl_bindings']
            return d

        def copy_sdef(s):
            """Copy schema definition structure with conditional parts for cloning."""
            d = prune_parts(s.prejson())
            if 'tables' in d:
                del d['tables']
            return d

        def copy_tdef_core(t):
            """Copy table definition structure with conditional parts excluding fkeys."""
            d = prune_parts(t.prejson())
            d['column_definitions'] = [ prune_parts(c) for c in d['column_definitions'] ]
            d['keys'] = [ prune_parts(c) for c in d.get('keys', []) ]
            if 'foreign_keys' in d:
                del d['foreign_keys']
            if 'annotations' not in d:
                d['annotations'] = {}
                d['annotations'][_clone_state_url] = 1 if copy_data else None
            return d

        def copy_tdef_fkeys(t):
            """Copy table fkeys structure."""
            def check(fkdef):
                for fkc in fkdef['referenced_columns']:
                    if fkc['schema_name'] == 'public' \
                       and fkc['table_name'] in {'ERMrest_Client', 'ERMrest_Group'} \
                       and fkc['column_name'] == 'RID':
                        raise ValueError("Cannot clone catalog with foreign key reference to %(schema_name)s:%(table_name)s:%(column_name)s" % fkc)
                return fkdef
            return [ prune_parts(check(d)) for d in t.prejson().get('foreign_keys', []) ]

        def copy_cdef(c):
            """Copy column definition with conditional parts."""
            return (sname, tname, prune_parts(c.prejson()))

        def check_column_compatibility(src, dst):
            """Check compatibility of source and destination column definitions."""
            def error(fieldname, sv, dv):
                return ValueError("Source/dest column %s mismatch %s != %s for %s:%s:%s" % (
                    fieldname,
                    sv, dv,
                    src.sname, src.tname, src.name
                ))
            if src.type.typename != dst.type.typename:
                raise error("type", src.type.typename, dst.type.typename)
            if src.nullok != dst.nullok:
                raise error("nullok", src.nullok, dst.nullok)
            if src.default != dst.default:
                raise error("default", src.default, dst.default)

        def copy_kdef(k):
            return (sname, tname, prune_parts(k.prejson()))

        for sname, schema in src_model.schemas.items():
            if sname in exclude_schemas:
                continue
            if sname not in dst_model.schemas:
                new_model.append(copy_sdef(schema))

            for tname, table in schema.tables.items():
                if table.kind != 'table':
                    logging.warning('Skipping cloning of %s %s:%s' % (table.kind, sname, tname))
                    continue

                if 'RID' not in table.column_definitions.elements:
                    raise ValueError("Source table %s.%s lacks system-columns and cannot be cloned." % (sname, tname))

                if sname not in dst_model.schemas or tname not in dst_model.schemas[sname].tables:
                    new_model.append(copy_tdef_core(table))
                    clone_states[(sname, tname)] = 1 if copy_data else None
                    fkeys_deferred[(sname, tname)] = copy_tdef_fkeys(table)
                else:
                    src_columns = { c.name: c for c in table.column_definitions }
                    dst_columns = { c.name: c for c in dst_model.schemas[sname].tables[tname].column_definitions }

                    for cname in src_columns:
                        if cname not in dst_columns:
                            new_columns.append(copy_cdef(src_columns[cname]))
                        else:
                            check_column_compatibility(src_columns[cname], dst_columns[cname])

                    for cname in dst_columns:
                        if cname not in src_columns:
                            raise ValueError("Destination column %s.%s.%s does not exist in source catalog." % (sname, tname, cname))

                    src_keys = { tuple(sorted(c.name for c in key.unique_columns)): key for key in table.keys }
                    dst_keys = { tuple(sorted(c.name for c in key.unique_columns)): key for key in dst_model.schemas[sname].tables[tname].keys }

                    for utuple in src_keys:
                        if utuple not in dst_keys:
                            new_keys.append(copy_kdef(src_keys[utuple]))

                    for utuple in dst_keys:
                        if utuple not in src_keys:
                            raise ValueError("Destination key %s.%s(%s) does not exist in source catalog." % (sname, tname, ', '.join(utuple)))

                    clone_states[(sname, tname)] = dst_model.schemas[sname].tables[tname].annotations.get(_clone_state_url)
                    if dst_model.schemas[sname].tables[tname].foreign_keys:
                        # assume that presence of any destination foreign keys means we already completed
                        return dst_catalog
                    else:
                        fkeys_deferred[(sname, tname)] = copy_tdef_fkeys(table)

        # apply the stage 1 model to the destination in bulk
        if new_model:
            dst_catalog.post("/schema", json=new_model).raise_for_status()

        for sname, tname, cdef in new_columns:
            dst_catalog.post("/schema/%s/table/%s/column" % (urlquote(sname), urlquote(tname)), json=cdef).raise_for_status()

        for sname, tname, kdef in new_keys:
            dst_catalog.post("/schema/%s/table/%s/key" % (urlquote(sname), urlquote(tname)), json=kdef).raise_for_status()

        # copy data in stage 2
        if copy_data:
            page_size = 10000
            for sname, tname in clone_states.keys():
                tname_uri = "%s:%s" % (urlquote(sname), urlquote(tname))
                if clone_states[(sname, tname)] == 1:
                    # determine current position in (partial?) copy
                    r = dst_catalog.get("/entity/%s@sort(RID::desc::)?limit=1" % tname_uri).json()
                    if r:
                        last = r[0]['RID']
                    else:
                        last = None

                    while True:
                        page = self.get(
                            "/entity/%s@sort(RID)%s?limit=%d" % (
                                tname_uri,
                                ("@after(%s)" % urlquote(last)) if last is not None else "",
                                page_size
                            )
                        ).json()
                        if page:
                            dst_catalog.post("/entity/%s?nondefaults=RID,RCT,RCB" % tname_uri, json=page)
                            last = page[-1]['RID']
                        else:
                            break

                    # record our progress on catalog in case we fail part way through
                    dst_catalog.put(
                        "/schema/%s/table/%s/annotation/%s" % (
                            urlquote(sname),
                            urlquote(tname),
                            urlquote(_clone_state_url),
                        ),
                        json=2
                    )
                elif clone_states[(sname, tname)] is None and (sname, tname) in {
                        ('public', 'ERMrest_Client'),
                        ('public', 'ERMrest_Group'),
                }:
                    # special sync behavior for magic ermrest tables
                    # HACK: these are assumed small enough to join via local merge of arrays
                    want = sorted(self.get("/entity/%s?limit=none" % tname_uri).json(), key=lambda r: r['ID'])
                    have = sorted(dst_catalog.get("/entity/%s?limit=none" % tname_uri).json(), key=lambda r: r['ID'])
                    create = []
                    update = []

                    pos_want = 0
                    pos_have = 0
                    while pos_want < len(want):
                        while pos_have < len(have) and have[pos_have]['ID'] < want[pos_want]['ID']:
                            # dst-only rows will be retained as is
                            pos_have += 1
                        if pos_have >= len(have) or have[pos_have]['ID'] > want[pos_want]['ID']:
                            # src-only rows will be inserted
                            create.append(want[pos_want])
                            pos_want += 1
                        else:
                            # overlapping rows will be updated
                            update.append(want[pos_want])
                            pos_want += 1

                    dst_catalog.post("/entity/%s?nondefaults=RCT,RCB" % tname_uri, json=create)
                    dst_catalog.put(
                        "/attributegroup/%s/ID;%s" % (
                            tname_uri,
                            ",".join([
                                urlquote(c.name)
                                for c in src_model.schemas[sname].tables[tname].column_definitions
                                if c.name not in {'RID', 'RMT', 'RMB', 'ID'}
                            ])
                        ),
                        json=update
                    )

                    # record our progress on catalog in case we fail part way through
                    dst_catalog.put(
                        "/schema/%s/table/%s/annotation/%s" % (
                            urlquote(sname),
                            urlquote(tname),
                            urlquote(_clone_state_url),
                        ),
                        json=2
                    )

        # apply stage 2 model in bulk only... we won't get here unless preceding succeeded
        new_fkeys = []
        for fkeys in fkeys_deferred.values():
            new_fkeys.extend(fkeys)

        if new_fkeys:
            dst_catalog.post("/schema", json=new_fkeys)

        # truncate cloning history
        if truncate_after:
            snaptime = dst_catalog.get("/").json()["snaptime"]
            dst_catalog.delete("/history/,%s" % urlquote(snaptime))

        return dst_catalog

class ErmrestSnapshot(ErmrestCatalog):
    """Persistent handle for an ERMrest catalog snapshot.

    Inherits from ErmrestCatalog and provides the same interfaces,
    except that the interfaces are now bound to a fixed snapshot
    of the catalog.
    """
    def __init__(self, scheme, server, catalog_id, snaptime, credentials=None, caching=True, session_config=None):
        """Create ERMrest catalog snapshot binding.

           Arguments:
             scheme: 'http' or 'https'
             server: server FQDN string
             catalog_id: e.g., '1'
             snaptime: e.g., '2PM-DGYP-56Z4'
             credentials: credential secrets, e.g. cookie
             caching: whether to retain a GET response cache
        """
        super(ErmrestSnapshot, self).__init__(scheme, server, catalog_id, credentials, caching, session_config)
        self._server_uri = "%s@%s" % (
            self._server_uri,
            snaptime
        )
        self._snaptime = snaptime

    @property
    def snaptime(self):
        """The snaptime for this catalog snapshot instance."""
        return self._snaptime

    def _pre_mutate(self, path, headers, guard_response=None):
        """Override and disable mutation operations.

        When called by the super-class, this method raises an exception.
        """
        raise ErmrestCatalogMutationError('Catalog snapshot is immutable')



__version__ = "1.3.0"

from deriva.core.utils.core_utils import *
from deriva.core.base_cli import BaseCLI, KeyValuePairArgs
from deriva.core.deriva_binding import DerivaBinding, DerivaPathError, DerivaClientContext
from deriva.core.deriva_server import DerivaServer
from deriva.core.ermrest_catalog import ErmrestCatalog, ErmrestSnapshot, ErmrestCatalogMutationError
from deriva.core.polling_ermrest_catalog import PollingErmrestCatalog
from deriva.core.hatrac_store import HatracStore, HatracHashMismatch, HatracJobPaused, HatracJobAborted, \
    HatracJobTimeout



class DerivaDownloadError(RuntimeError):
    pass


class DerivaDownloadConfigurationError(ValueError):
    pass


class DerivaDownloadAuthenticationError(Exception):
    pass


class DerivaDownloadAuthorizationError(Exception):
    pass



from xml.dom.minidom import getDOMImplementation
from deriva.core import DerivaServer, get_credential, BaseCLI
import time
import sys


class SitemapBuilder:

    """Class to build sitemaps from deriva catalogs.

      - Typical usage is:

         1. Create a SiteMapBuilder class
         2. For each table to include in the sitemap, create a table spec and populate it with the
            table's data. Typically, the creating and populating for a table is done by a single
            call to add_table_spec.
         3. Call write_sitemap to write out the site map

      - In the simplest case, this would create a sitemap with two tables::

            sb = SitemapBuilder("https", "myhost.org", 1)
            sb.add_table_spec("schema1", "table1")
            sb.add_table_spec("schema2", "table2")
            sb.write_sitemap(sys.stdout)

      - If you want to include only a subset of rows of a table, you can pass a datapath to add_table_spec::

           pb = catalog.getPathBuilder()
           path=pb.schema3.table3.filter(pb.schema3.table3.Species=="Homo sapiens")
           sb.add_table_spec("schema3", "table3", datapath=path)

      - If you want to do something more customizable, you can populate the spec yourself::

           rows = do_something_complicated()
           spec = sb.add_table_spec("schema4", "table4", populate=False)
           sb.set_table_spec_data(rows)
           sb.add_fkey_times(spec)

        Note: if add_table_spec populates the spec for you, it will set the modification time based on
        the times in the table and on all single-valued incoming foreign keys (because those are the
        most likely to affect a Chaise record page). If you're populating the spec yourself (i.e., if
        you called add_table_spec with populate=False), you can call add_fkey_times to make those time
        adjustments.

       Limitations:
         Sitemaps should be no more than 50MB in size and should contain no more
         than 50,000 URLS (https://www.sitemaps.org/faq.html#faq_sitemap_size),
         but this class doesn't enforce those limits. (Note: A site can have
         multiple sitemaps).

         A URL element should have no more than 1000 images associated with it
         (https://support.google.com/webmasters/answer/178636?hl=en), but this
         class doesn't enforce that limit.

         This class assumes that all images in a catalog will have the same
         license.
    """

    def __init__(self, protocol, host, catalog_id, license_url=None):
        """ Creates the sitemap object
        :param protocl: protocol (e.g., https)
        :param host: the host name
        :param catalog_id: the catalog id
        :param license_url: the url to the license used for images in this catalog
        """

        if sys.version_info[0] < 3:
            raise NotImplementedError("SitemapBuilder is only supported for Python 3.x")
        self.protocol = protocol
        self.host = host
        self.catalog_id = catalog_id
        self.license_url = license_url
        self._create_xml_tree()
        # Don't authenticate - we want only publicly-available content in the index
        self.server = DerivaServer(protocol, host)
        self.catalog = self.server.connect_ermrest(catalog_id)
        self.model = self.catalog.getCatalogModel()
        self.pb = self.catalog.getPathBuilder()
        self.table_specs = []

    def add_table_spec(self, schema, table, datapath=None, populate=True, priority=None):
        """Create a table spec and add it to the sitemap.

        :param schema: the name of the table's schema
        :param table: the table name
        :param datapath: a datapath to use to populate the table (regardless of the value
                         of "populate")
        :param populate: an indication of whether or not the spec should be populated.
                         If populate==True, the table spec's data will be populated with the results
                         from the query specified by "datapath" (if non-None) or all the rows of the table.
                         If populate==False (and datapath is not set), the table spec's data will
                         need to be set via a call to set_table_spec_data()
        :param priority: the priority to assign to sitemap entries created from this table
                         (see https://www.sitemaps.org/protocol.html for a discussion of priorities)

        """
        spec = {
            "table": self.get_table(schema, table),
            "incoming_fkeys": self.model.table(schema, table).referenced_by,
            "url":
            "{protocol}://{host}/chaise/record/?{catalog_id}/{schema}:{table}/RID="
            .format(
                protocol=self.protocol,
                host=self.host,
                catalog_id=self.catalog_id,
                schema=schema,
                table=table)
        }
        if priority is not None:
            spec["priority"] = str(priority)

        if populate or datapath is not None:
            self.populate_spec(spec, datapath)

        self.table_specs.append(spec)
        return spec

    def write_sitemap(self, file):
        """Write out the sitemap

           :param file: The file to write to
        """

        for spec in self.table_specs:
            self._add_spec_records(spec)

        self.doc.writexml(file, encoding="UTF-8", newl="\n", addindent="  ")

    def _create_xml_tree(self):
        """Create empty xml tree"""
        impl = getDOMImplementation()
        self.doc = impl.createDocument(
            "http://www.sitemaps.org/schemas/sitemap/0.9", "urlset", None)
        self.tree = self.doc.documentElement
        self.tree.setAttributeNS("xmlns", "xsi",
                                 "http://www.w3.org/2001/XMLSchema-instance")
        self.tree.setAttributeNS(
            "xsi", "schemaLocation",
            "http://www.sitemaps.org/schemas/sitemap/0.9 http://www.sitemaps.org/schemas/sitemap/0.9/sitemap.xsd")
        self.tree.setAttribute(
            "xmlns:image", "http://www.google.com/schemas/sitemap-image/1.1")
        self.tree.setAttribute("xmlns", "http://www.sitemaps.org/schemas/sitemap/0.9")

    def _add_spec_records(self, spec):
        """Add records to xml doc"""
        rows = spec.get("data")
        basepath = spec.get("url")
        priority = spec.get("priority")

        for rec in rows:
            url_node = self.doc.createElement("url")
            loc_node = self.doc.createElement("loc")
            loc_node.appendChild(
                self.doc.createTextNode(
                    "{base}{rid}".format(base=basepath, rid=rec.get("RID"))))
            url_node.appendChild(loc_node)
            lastmod_node = self.doc.createElement("lastmod")
            lastmod_node.appendChild(self.doc.createTextNode(rec.get("RMT")))
            url_node.appendChild(lastmod_node)
            if priority is not None:
                priority_node = self.doc.createElement("priority")
                priority_node.appendChild(self.doc.createTextNode(priority))
                url_node.appendChild(priority_node)
            images = rec.get("images")
            # the default dom implementation doesn't include namespaces in output,
            # so use "image:" as part of the element name
            if images is not None:
                for image in images:
                    if image.get("image_url") is not None:
                        image_node = self.doc.createElementNS(
                            "http://www.google.com/schemas/sitemap-image/1.1",
                            "image:image")
                        image_loc_node = self.doc.createElementNS(
                            "http://www.google.com/schemas/sitemap-image/1.1",
                            "image:loc")
                        image_loc_node.appendChild(
                            self.doc.createTextNode(image.get("image_url")))
                        image_node.appendChild(image_loc_node)
                        if image.get("image_caption") is not None:
                            image_caption_node = self.doc.createElementNS(
                                "http://www.google.com/schemas/sitemap-image/1.1",
                                "image:caption")
                            image_caption_node.appendChild(
                                self.doc.createTextNode(image.get("image_caption")))
                            image_node.appendChild(image_caption_node)
                        if image.get("image_title") is not None:
                            image_title_node = self.doc.createElementNS(
                                "http://www.google.com/schemas/sitemap-image/1.1",
                                "image:title")
                            image_title_node.appendChild(
                                self.doc.createTextNode(image.get("image_title")))
                            image_node.appendChild(image_title_node)
                        if self.license_url is not None:
                            image_license_node = self.doc.createElementNS(
                                "http://www.google.com/schemas/sitemap-image/1.1",
                                "image:license")
                            image_license_node.appendChild(
                                self.doc.createTextNode(self.license_url))
                            image_node.appendChild(image_license_node)
                        url_node.appendChild(image_node)
            self.tree.appendChild(url_node)

    def populate_spec(self, spec, datapath):
        """Add data to a spec

        :param spec: the spec to populate
        :param datapath: optional datapth to use (if None, all records
                         from the spec table will be incuded)
        """
        if datapath is None:
            rows = spec["table"].entities()
        else:
            rows = datapath.entities()

        spec["data"] = rows
        self.add_fkey_times(spec)

    def add_fkey_times(self, spec):
        """Replace the RMT in each row with the greatest value of
           the row's RMT and the RMTs of all single-column-fkey-linked
           tables.

           :param spec: a populated spec
        """
        rows = spec["data"]
        ref_table = spec["table"]
        fkeys = spec["incoming_fkeys"]
        rdict = dict()
        if len(fkeys) > 0:
            for r in rows:
                rdict[r.get("RID")] = {"row": r}
        for fkey in fkeys:
            if len(fkey.referenced_columns) == 1:
                # Assume that if a related table is visible in chaise,
                # a single-valued fkey exists.
                self._apply_fkey(ref_table, rdict, fkey)

    def _apply_fkey(self, ref_table, rdict, fkey):
        fkc = fkey.foreign_key_columns[0]
        rc = fkey.referenced_columns[0]
        fk_table = self.get_table(fkc["schema_name"], fkc["table_name"])
        fk_col = self.get_column(fk_table, fkc["column_name"])
        ref_col = self.get_column(ref_table, rc["column_name"])
        datapath = ref_table.alias("ref")\
            .link(fk_table.alias("fk"), on=(ref_col == fk_col))
        fk_rows = datapath.attributes(datapath.ref.RID, datapath.fk.RMT)
        for fk_row in fk_rows:
            entry = rdict.get(fk_row["RID"])
            if entry is not None:
                fk_time = self.ermrest_time_to_float(fk_row["RMT"])
                row = entry.get("row")
                entry_time = entry.get("rmt_float")
                if entry_time is None:
                    entry_time = self.ermrest_time_to_float(row["RMT"])
                if fk_time > entry_time:
                    row["RMT"] = fk_row["RMT"]
                    entry["rmt_float"] = fk_time
                else:
                    entry["rmt_float"] = entry_time

    @staticmethod
    def ermrest_time_to_float(ermrest_time):
        """Converts a time string as returned by ermrest to a floating-point number
        """
        return(time.mktime(time.strptime(ermrest_time, "%Y-%m-%dT%H:%M:%S.%f%z")))

    def get_table(self, sname, tname):
        """Gets a datapath table
        """
        return self.pb.schemas.get(sname).tables.get(tname)

    def get_column(self, table, column_name):
        """Gets a datapath column
        """
        return table.column_definitions.get(column_name)

    def set_table_spec_data(self, spec, rows):
        """Add data to a table spec. This should be used if you want to associate
           image data with the rows of this table, or if you want to include
           only a subset of the table's rows in the sitemap. Otherwise, you can
           just use populate=True in your call to add_table_spec().

        :param spec: a table spec returned by add_table_spec()
        :param rows: an array of dictionary objects corresponding to the rows
           of the table. Each row MUST have RID and RMT elements and MAY have
           an "images" element. If "images" is present, each entry should have:
              "image_url" : the absolute URL of the image (jpeg, etc.)
              "image_caption" : a caption for the image (optional)
              "image_title" : a title for the image (optional)

           for example::

              [
                 {
                    "RID": '1-2345',
                    "RMT": '018-12-03T18:29:38.348231-08:00'},
                 {
                    "RID": '1-2346',
                    "RMT": '018-12-03T18:29:38.348231-08:00',
                    "images": [
                        {
                           "image_url": "https://myhost.org/hatrac/images/image1.jpg",
                           "image_title": "MyProject Image 2-3456: An Awesome Image",
                           "image_caption": "This is an awesome image"
                        },
                        {
                           "image_url": "https://myhost.org/hatrac/images/image2.jpg",
                        }
                    ]
                 }
              ]
        """
        spec["data"] = rows



from deriva.transfer import DerivaDownloadError, DerivaDownloadAuthorizationError, DerivaDownloadAuthenticationError, \
    DerivaDownloadConfigurationError


class DerivaBackupError(DerivaDownloadError):
    pass


class DerivaBackupConfigurationError(DerivaDownloadConfigurationError):
    pass


class DerivaBackupAuthenticationError(DerivaDownloadAuthenticationError):
    pass


class DerivaBackupAuthorizationError(DerivaDownloadAuthorizationError):
    pass



from copy import deepcopy
"""Utility functions that really belong in webauthn, but that are here to avoid having deriva-py depend on webauthn"""


def get_wallet_entries(wallet, credential_type="oauth2", **kwargs):
    """
    :param wallet: wallet - the wallet to examine (from a Client object)
    :param credential_type: the type of credential requested from the wallet
    :param kwargs: keyword arguments
    :return: a list of oauth2 credentials obtained from auth.globus.org with the requested scope

    Currently, only "oauth2" is supported as a credential type, and the following keyword args are supported
    (all of these are optional):

        * `credential_source` - where the credentials came from (e.g., "https://auth.globus.org").
        * `resource_server` - the resource server associated with a credential.
        * `scopes` - a list of desired scopes.

    Example #1:
    ::
        wallet = client.wallet.extra_values.get("wallet")
        entries = get_wallet_entries(wallet, "oauth2", resource_server="identifiers.globus.org")

    :return: a list of oauth2 credentials associated with the resource server `identifiers.globus.org`.

    Example #2:
    ::
        entries = get_wallet_entries(wallet, "oauth2", credential_source="https://auth.globus.org")

    :return: a list of oauth2 credentials obtained from `auth.globus.org`.

    Example #3:
    ::
        entries = get_wallet_entries(
                      wallet,
                      "oauth2",
                      credential_source="https://auth.globus.org",
                      scopes=["https://auth.globus.org/scopes/identifiers.globus.org/create_update")

    :return: a list of oauth2 credentials obtained from `auth.globus.org` with the requested scope.

    .. note::
        Eventually, we may support wallets with multiple credential types (at the same time) for talking
        to a variety of remote servers, at which point the implementation will probably become a lot more
        generalized (with registered wallet providers, etc.).

    """
    if wallet is None:
        return []
    wallet = deepcopy(wallet)

    matching_entries = []
    
    # wallet structure is
    # {
    #   credential_type : {
    #     credential_source : [
    #       <oauth entry, w/ "scope", "resource_server", "access_token", etc.>
    
    if credential_type != "oauth2":
        raise NotImplementedError("Only 'oauth2' wallet credentials are supported")
    wallet = wallet.get(credential_type)
    if wallet is None:
        return []

    # wallet structure is now
    #   {
    #     credential_source : [
    #       <oauth entry, w/ "scope", "resource_server", "access_token", etc.>
    
    if kwargs.get('credential_source') is not None:
        matching_entries = wallet.get(kwargs.get('credential_source'))
        if matching_entries is None:
            return []
    else:
        for entries in wallet.values():
            matching_entries = matching_entries + entries

    resource_server = kwargs.get('resource_server')
    to_remove = []    
    if resource_server is not None:
        for entry in matching_entries:
            if entry.get('resource_server') != resource_server:
                to_remove.append(entry)

    for entry in to_remove:
        matching_entries.remove(entry)

    to_remove = []
    if kwargs.get('scopes') is not None:
        for entry in matching_entries:
            if entry.get("scope") is not None:
                entry_scopes = entry.get("scope").split()
                for a_scope in kwargs.get('scopes'):
                    if a_scope not in entry_scopes:
                        to_remove.append(entry)
                        break

    for entry in to_remove:
        matching_entries.remove(entry)
                    
    return matching_entries
